""" xUnit.net console test runner

Input files could be list of pairs of assembly file and config file.
Configuration files must end in .config

Options:

    * parallel : str, None
               : Set parallelization based on option
               : Available values:
               :   none : turn off all parallelization
               :   collections : only parallelize collections
               :   assemblies : only parallelize assemblies
               :   all : parallelize assemblies & collection

    * max_threads : int, None
                  : Maximum thread count for collection parallelization
                  :   0 : run with unbounded thread count
                  :   >0 : limit task thread pool size to 'count'

    * no_shadow : bool, None
                : Do not shadow copy assemblies

    * team_city : bool, None
                : Forces TeamCity mode (normally auto-detected)

    * appveyor : bool, None
               : Forces AppVeyor CI mode (normally auto-detected)

    * no_logo : bool, None
              : Do not show the copyright message

    * quiet : bool, None
            : Do not show progress messages

    * serialize : bool, None
                : Serialize all test cases (for diagnostic purposes only

    * traits : dict, None
             : Only run tests with matching name/value traits
             : If specified more than once, acts as an OR operation

    * no_traits : dict, None
                : Do not run tests with matching name/value traits
                : If specified more than once, acts as an AND operation

    * methods : list, None
              : Run a given test method (should be fully specified;
              : i.e., 'MyNamespace.MyClass.MyTestMethod')
              : If specified more than once, acts as an OR operation

    * classes : list, None
              : Run all methods in a given test class (should be fully
              : specified; i.e., 'MyNamespace.MyClass')
              : If specified more than once, acts as an OR operation

    * reports : dict, None
              : Pairs of output style and filename.
              : Available output styles:
              :   xml : output results to xUnit.net v2 style XML file
              :   xmlv1 : output results to xUnit.net v1 style XML file
              :   nunit : output results to NUnit-style XML file
              :   html : output results to HTML file

Requirements:

    * xunit
      to install, for example run `choco install xunit`

"""
import os
from pybuildtool import BaseTask, expand_wildcard, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        # Set parallelization
        c = cfg.get('parallel')
        if c is not None:
            args.append('-parallel')
            args.append(c)

        # Maximum thread count
        c = cfg.get('max_threads')
        if c is not None:
            args.append('-maxthreads')
            args.append(c)

        # Shadow copy
        c = cfg.get('no_shadow')
        if c:
            args.append('-noshadow')

        # TeamCity mode
        c = cfg.get('teamcity')
        if c:
            args.append('-teamcity')

        # AppVeyor CI mode
        c = cfg.get('appveyor')
        if c:
            args.append('-appveyor')

        # No logo
        c = cfg.get('no_logo')
        if c:
            args.append('-nologo')

        # Quiet
        c = cfg.get('quiet')
        if c:
            args.append('-quiet')

        # Serialize
        c = cfg.get('serialize')
        if c:
            args.append('-serialize')

        # Trait
        c = cfg.get('traits', {})
        for key, value in c.items():
            args.append('-trait')
            args.append('%s=%s' % (key, value))

        # No trait
        c = cfg.get('no_traits', {})
        for key, value in c.items():
            args.append('-notrait')
            args.append('%s=%s' % (key, value))

        # Methods
        for method in make_list(cfg.get('methods')):
            args.append('-method')
            args.append(method)

        # Classes
        for cls in make_list(cfg.get('classes')):
            args.append('-class')
            args.append(cls)

        # Reports
        c = cfg.get('reports', {})
        for key, value in c.items():
            if key not in ('xml', 'xmlvi', 'nunit', 'html'):
                self.bld.fatal('Unknown xunit report style: %s.' % key)

            args.append('-' + key)
            args.append(os.path.realpath(expand_wildcard(self.group, value)))


    def perform(self):
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
                exe=executable,
                arg=' '.join(self.args),
                in_=' '.join(self.file_in),
            ))


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] =\
            conf.find_program('xunit.console')[0]



"""
PostCSS is a tool for transforming styles with JS plugins. These plugins can
lint your CSS, support variables and mixins, transpile future CSS syntax,
inline images, and more.

https://github.com/postcss/postcss


Options:

    * map : str, None
          : Create an external sourcemap

    * no_map : bool, None
             : Disable the default inline sourcemaps

    * verbose : bool, None
              : Be verbose

    * env : str, None
          : A shortcut for setting NODE_ENV

    * include_dotfiles : bool, None
                       : Enables glob to match files/dirs that begin with "."

    * use : list, None
          : List of postcss plugins to use

    * parser : str, None
             : Custom postcss parser

    * stringifier : str, None
                  : Custom postcss stringifier

    * syntax : str, None
             : Custom postcss syntax


Requirements:

    * node.js
    * postcss-cli
      to install, run `npm install --save-dev postcss-cli`

"""

import os
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        self.add_bool_args('no_map', 'include_dotfiles')

        self.add_path_args('map')

        self.add_list_args_multi('use')

        self.add_str_args('env', 'parser', 'stringifier', 'syntax')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} -o {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    bin_path = 'node_modules/postcss-cli/bin/postcss'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('postcss')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Less-css is a CSS pre-processor, meaning that it extends the CSS language,
adding features that allow variabels, and mixins, functions and many other
techniques that allow you to make CSS that is more maintainable, themable and
extendable.

Options:

    * keep_line_breaks     : bool, False, keep line breaks

Requirements:

    * node.js
    * less
      to install, run `npm install --save-dev less`

"""

import os
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_replace_patterns_': ((r'\.less$', '.css'),)
    }
    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        # keep line breaks
        if cfg.get('keep_line_breaks', False):
            args.append('--keep-line-breaks')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    bin_path = 'node_modules/less/bin/lessc'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('lessc')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
HTML5 Linter based on Google Style Guide

Options:

    * disable : list:str, [], any combination of: doctype, entities,
                trailing_whitespace, tabs, charset, void_element,
                optional_tag, type_attribute, concerns_separation, protocol,
                names, capitalization, quotation, indentation, formatting,
                boolean_attribute, invalid_attribute, void_zero,
                invalid_handler, http_equiv, extra_whitespace

Requirements:

    * html-linter
      to install, run `pip install html-linter`

"""
from pybuildtool import BaseTask, make_list

tool_name = __name__

class Task(BaseTask):

    DISABLE_LIST = set(('doctype', 'entities', 'trailing_whitespace', 'tabs',
            'charset', 'void_element', 'optional_tag', 'type_attribute',
            'concerns_separation', 'protocol', 'names', 'capitalization',
            'quotation', 'indentation', 'formatting', 'boolean_attribute',
            'invalid_attribute', 'void_zero', 'invalid_handler', 'http_equiv',
            'extra_whitespace'))

    conf = {
        '_source_grouped_': True,
    }
    name = tool_name

    def prepare(self):
        cfg = self. conf
        args = self.args

        c = make_list(cfg.get('disable'))
        if c:
            invalid_disable_items = set(c) - self.DISABLE_LIST
            if invalid_disable_items:
                self.bld.fatal('invalid disable configuration items: ' +\
                        ', '.join(invalid_disable_items))
            args.append('--disable=' + ','.join(c))


    def perform(self):
        executable = self.env['%s_BIN' % tool_name.upper()]
        for file_in in self.file_in:
            return_code = self.exec_command(
                '{exe} {arg} {in_}'.format(
                exe=executable,
                arg=' '.join(self.args),
                in_=file_in,
            ))
            if return_code:
                print('Found syntax errors in %s\n' % file_in)
                return 1
        return 0


def configure(conf):
    bin_path = conf.find_program('html_lint.py')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Sphinx is a tool that makes it easy to create intelligent and beautiful
documentation, written by Georg Brandl and licensed under the BSD license.

It was originally created for the Python documentation, and it has excellent
facilities for the documentation of sofware projects in a range of languages.

Options:

    * builder    : str, html
                   Builder to use; default is html.
    * source_dir : str, None, required
                   Source directory.
    * output_dir : str, None, required
                   Output directory.
    * temp_dir   : str, None
                   Path for the cached environment and doctree files
                   (default: output_dir/.doctrees)
    * conf_dir   : str, None
                   Path where configuration file (conf.py) is located
                   (default: same as sourcedir).
    * settings   : dict, {}
                   Override setting in configuration file.
    * context    : dict, {}
                   Pass a value into HTML templates.
    * jobs       : int, None
                   Build in parallel with N processes where possible.

Requirements:

    * sphinx
      to install, run `pip install sphinx`

"""
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)
            if self.workdir is None:
                self.bld.fatal(cfg['work_dir'] + ' not found.')

        source_dir = expand_resource(self.group, cfg['source_dir'])
        if source_dir is None:
            self.bld.fatal(cfg['source_dir'] + ' not found.')
        args.append(source_dir)

        output_dir = expand_resource(self.group, cfg['output_dir'])
        if output_dir is None:
            self.bld.fatal(cfg['output_dir'] + ' not found.')
        args.append(output_dir)

        c = cfg.get('temp_dir')
        if c:
            temp_dir = expand_resource(self.group, c)
            if temp_dir is None:
                self.bld.fatal(c + ' not found.')
            args.append('-d ' + temp_dir)

        c = cfg.get('conf_dir')
        if c:
            conf_dir = expand_resource(self.group, c)
            if conf_dir is None:
                self.bld.fatal(c + ' not found.')
            args.append('-c ' + conf_dir)

        c = cfg.get('builder', 'html')
        args.append('-b ' + c)

        c = cfg.get('jobs')
        if c:
            args.append('-j %i' % c)

        c = cfg.get('settings', {})
        for key, value in c.items():
            args.append('-D %s=%s' % (key, value))

        c = cfg.get('context', {})
        for key, value in c.items():
            args.append('-A %s=%s' % (key, value))


    def perform(self):
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['SPHINX_BUILD_BIN']
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
                exe=executable,
                arg=' '.join(self.args),
                in_=' '.join(self.file_in),
            ),
            **kwargs)


def configure(conf):
    conf.env['SPHINX_BUILD_BIN'] = conf.find_program('sphinx-build')[0]



"""
CMake link commands for ROS packages. The lint commands perform static
checking of Python or C++ source code for errors and standards compliance.
See: http://wiki.ros.org/roslint

Options:

    * work_dir    : str, None, Change current directory
    * output      : str, None, By default, the output is formatted to ease
                    emacs parsing. Visual Studio compatible output (vs7) may
                    also be used. Other formats are unsupported.
                    Possible values: vs7
    * filter      : list, []
                    Specify category-filters to apply: only error messages
                    whose category names pass the filters will be printed.
                    Category names are printed with the message and look like
                    "[whitespace/indent]".
                    Filter are evaluated left to right.
                    "-FOO" and "FOO" means "do not print categories that start
                    with FOO"
                    "+FOO" means "do print categories that start with FOO".
    * counting    : str, None
                    The total number of errors found is always printed. If
                    'top level' is provided, then the count of errors in each
                    of the top-level categories like 'build' and 'whitespace'
                    will also be printed. If 'detailed' is provided, then a
                    count is provided for each category like 'build/class'.
                    Possible values: total, toplevel, detailed
    * root_dir    : str, None
                    The root directory used for deriving header guard CPP
                    variable. By default, the header guard CPP variable is
                    calculated as the relative path to the directory that
                    contains .git, .hg, or .svn. When this flag is specified,
                    the relative path is calculated from the specified
                    directory. If the specified directory does not exist, this
                    flag is ignored.
    * line_length : int, None
                    This is the allowed line length for the project. The
                    default value is 80 characters.
    * extensions  : list, ['c', 'cpp', 'h']
                    The allowed file extensions that cpplint will check.

Requirements:

    * roslint
      to install, ??

"""
import os
from pybuildtool import BaseTask, expand_resource, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        # Change current directory
        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)
            if self.workdir is None:
                self.bld.fatal(cfg['work_dir'] + ' not found.')

        # Output
        c = cfg.get('output')
        if c:
            args.append('--output=' + c)

        # Filters
        c = make_list(cfg.get('filter'))
        if c:
            args.append('--filter=' + ','.join(c))

        # Counting
        c = cfg.get('counting')
        if c:
            args.append('--counting=' + c)

        # Root
        c = cfg.get('root_dir')
        if c:
            args.append('--root=' + expand_resource(self.group, c))

        # Line length
        c = cfg.get('line_length')
        if c:
            args.append('--linelength=%i' % c)

        # Extensions
        c = make_list(cfg.get('extensions', ['c', 'cpp', 'h']))
        args.append('--extensions=' + ','.join(c))


    def perform(self):
        if self.file_out:
            self.bld.fatal("%s doesn't produce files" % tool_name.capitalize())

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        # TODO: roslint doesn't work in python3, xrange and bytes
        return self.exec_command(
            'python2 {exe} {arg} {in_}'.format(
                exe=executable,
                arg=' '.join(self.args),
                in_=' '.join(self.file_in),
            ),
            **kwargs)


def configure(conf):
    bin_path = '/usr/libexec/roslint/cpplint'
    conf.start_msg("Checking for program '%s'" % 'cpplint')
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('cpplint')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



""" Implements watchdog.
"""
import os
import re
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

class FileChangeHandler(FileSystemEventHandler):

    app = None
    file_patterns = None

    def __init__(self, app):
        super(FileChangeHandler, self).__init__()
        self.app = app
        self.file_patterns = []


    def on_any_event(self, event):
        if event.is_directory:
            return

        if event.src_path == self.app.config_file:
            self.app.reload = True

        else:
            filters = self.file_patterns
            for name in event.src_path.split(os.path.sep):
                filters = self.filter_reduce(name, filters)

            filters = [f for f in filters if len(f) == 0 or f[0] != '**']
            if len(filters):
                self.app.rebuild = True


    def set_files(self, files):
        self.file_patterns = [self.parse_pattern(f) for f in files]


    @staticmethod
    def parse_pattern(x):
        filter_ = []
        x = x.replace('\\', '/').replace('//', '/')
        if x.endswith('/'):
            x += '**'

        for k in x.split('/'):
            if k == '**':
                filter_.append(k)
            else:
                k = k.replace('.', '[.]').replace('*','.*').replace('?', '.').\
                        replace('+', '\\+')

                k = '^%s$' % k
                exp = re.compile(k)
                filter_.append(exp)

        return filter_


    @staticmethod
    def filter_reduce(name, filters):
        ret_filters = []
        for lst in filters:
            if not lst:
                pass
            elif lst[0] == '**':
                ret_filters.append(lst)
                if len(lst) > 1:
                    if lst[1].match(name):
                        ret_filters.append(lst[2:])
                else:
                    ret_filters.append([])
            elif lst[0].match(name):
                ret_filters.append(lst[1:])

        return ret_filters


class FileObserver(object):

    app = None
    observers = None
    handler = None

    def __init__(self, app):
        self.app = app
        self.observers = []
        self.handler = FileChangeHandler(app)


    def open(self, files):
        self.handler.set_files(files)

        dirnames = set()
        dirnames.add(os.path.dirname(self.app.config_file))

        for filename in files:
            while True:
                filename = os.path.dirname(os.path.realpath(filename))
                if os.path.exists(filename) or not filename:
                    break

            if filename:
                for dirname in dirnames:
                    if os.path.commonprefix([filename, dirname]) != dirname:
                        dirnames.add(filename)

        for dirname in dirnames:
            observer = Observer()
            observer.schedule(self.handler, dirname, recursive=True)
            observer.start()

            self.observers.append(observer)


    def close(self):
        for observer in self.observers:
            observer.stop()

        for observer in self.observers:
            observer.join()

        self.observers = []



"""
Preprocess CSS files using [Stylus](http://learnboost.github.io/stylus/).

Options:

    * work_dir : str, None
               : change current directory

    * use : list, None
          : utilize the Stylus plugin at <path>

    * inline : bool, None
             : use data URI

    * include : list, None
              : Add <path> to lookup paths

    * compress : bool, None
               : compress CSS output

    * firebug : bool, None
              : Emits debug infos in the generated CSS that can be used by the
              : FireStylus Firebug plugin

    * line_numbers : bool, None
                   : Emits comments in the generated CSS indicating the
                   : corresponding Stylus line

    * sourcemap : bool, None
                : Generates a sourcemap in sourcemaps v3 format

    * sourcemap_inline : bool, None
                       : Inlines sourcemap with full source text in base64
                       : format

    * sourcemap_root : str, None
                     : "sourceRoot" property of the generated sourcemap

    * sourcemap_base : str, None
                     : Base <path> from which sourcemap and all sources are
                     : relative

    * prefix : str, None
             : prefix all css classes

    * import : list, None
             : Import Stylus <file>

    * include_css : bool, None
                  : Include regular CSS on @import

    * deps : bool, None
           : Display dependencies of the compiled files

    * disable_cache : bool, None
                    : Disable caching

    * hoist_atrules : bool, None
                    : Move @import and @charset to the top

    * resolve_url : bool, None
                  : Resolve relative urls inside imports

    * resolve_url_nocheck : bool, None
                          : Like --resolve-url but without file existence check


Requirements:

    * node.js
    * stylus
      to install, run `npm install --save-dev stylus`

"""
import os
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    conf = {
        '_replace_patterns_': ((r'\.styl$', '.css'),)
    }
    workdir = None

    def prepare(self):
        cfg = self.conf
        self.args = ['--print']

        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)

        self.add_bool_args('inline', 'compress', 'firebug', 'line_numbers',
                'sourcemap', 'sourcemap_inline', 'include_css', 'deps',
                'disable_cache', 'hoist_atrules', 'resolve_url',
                'resolve_url_nocheck')

        self.add_path_args('sourcemap_base', opt_val_sep=' ')

        self.add_path_list_args_multi('use', 'include', 'import',
                opt_val_sep=' ')

        self.add_str_args('sourcemap_root', 'prefix', opt_val_sep=' ')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input, got %s' % (
                    tool_name.capitalize(), repr(self.file_in)))

        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output, got %s ' % (\
                    tool_name.capitalize(), repr(self.file_out)))

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} < {in_} > {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ),
        **kwargs)


def configure(conf):
    bin_path = 'node_modules/stylus/bin/stylus'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('stylus')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Clean-css is a fast and efficient Node.js library for minifying CSS files.

Options:

    * keep-line-breaks     : bool, False, keep line breaks
    * no-comments          : bool, False, remove all special comments,
                             i.e /*! comment */
    * first-special-comment: bool, False, remove all special comments,
                             but the first one
    * root-path            : str, None, a root path to which resolve absolute
                             @import rules and rebase relative URLs
    * skip-import          : bool, False, disable @import processing
    * skip_rebase          : bool, False, disable URLs rebasing
    * skip-advanced        : bool, False, disable advanced optimizations -
                             selector & property merging, reduction, etc
    * skip-aggressive-merging: bool, False, disable properties merging based
                               on their order
    * rounding-precision   : int, 2, rounding precision
    * compatibility        : str, None, [ie7, ie8], force compatibility mode
    * timeout              : int, 5, per connection timeout when fetching
                             remote @imports (in seconds)
    * debug                : shows debug information (minification time &
                             compression efficiency)

Requirements:

    * clean-css
      to install, run `npm install --save-dev clean-css-cli`
    * node.js

"""
import os
from shutil import copyfile, Error
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_replace_patterns_': ((r'\.css$', '.min.css'),),
    }
    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        # keep line breaks
        if cfg.get('keep-line-breaks', False):
            args.append('--keep-line-breaks')

        # remove all special comments, i.e /*! comment */
        if cfg.get('no-comments', False):
            args.append('--s0')

        # remove all special comments but the first one
        if cfg.get('first-special-comment', False):
            args.append('--s1')

        # a root path to which resolve absolute @import rules and rebase
        # relative URLs
        if cfg.get('root-path', None):
            args.append('--root=%s' % cfg['root-path'])

        # disable @import processing
        if cfg.get('skip-import', False):
            args.append('--skip-import')

        # disable URLs rebasing
        if cfg.get('skip_rebase', False):
            args.append('--skip-rebase')

        # disable advanced optimizations - selector & property merging,
        # reduction, etc
        if cfg.get('skip-advanced', False):
            args.append('--skip-advanced')

        # disable properties merging based on their order
        if cfg.get('skip-aggressive-merging', False):
            args.append('--skip-aggressive-merging')

        # rounding precision
        if cfg.get('rounding-precision', None):
            args.append('--rounding-precision=%s' % cfg['rounding-precision'])

        # force compatibility mode
        if cfg.get('compatibility', None):
            args.append('--compatibility=%s' % cfg['compatibility'])

        # per connection timeout when fetching remote @improts (in seconds)
        if cfg.get('timeout', None):
            args.append('--timeout=%s' % cfg['timeout'])

        # show debug information (minification time & compression efficiency)
        if cfg.get('debug', False):
            args.append('--debug')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        if not self.is_production():
            try:
                copyfile(self.file_in[0], self.file_out[0])
                return 0
            except (IOError, Error):
                self.bld.fatal('cannot copy file to ' + self.file_out[0])
            return -1

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} -o {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    bin_path = 'node_modules/clean-css-cli/bin/cleancss'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('cleancss')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Wrapper around libsass.

Options:

    * work_dir : str, None
               : change current directory

    * omit_source_map_url : bool, None
                          : Omit source map URL comment from output

    * indented_syntax : bool, None
                      : Treat data from stdin as sass code (versus scss)

    * output_style : str, None
                   : CSS output style (nested | expanded | compact | compressed)

    * indent_type : str, None
                  : Indent type for output CSS (space | tab)

    * indent_width : int, None
                   : Indent width; number of spaces or tabs (maximum value: 10)

    * linefeed : str, None
               : Linefeed style (cr | crlf | lf | lfcr)

    * source_comments : bool, None
                      : Include debug info in output

    * source_map : str, None
                 : Emit source map (boolean, or path to output .map file)

    * source_map_contents : bool, None
                          : Embed include contents in map

    * source_map_embed : bool, None
                       : Embed sourceMappingUrl as data URI

    * source_map_root : bool, None
                      : Base path, will be emitted in source-map as is

    * include_path : str, None
                   : Path to look for imported files

    * follow : bool, None
             : Follow symlinked directories

    * precision : int, None
                : The amount of precision allowed in decimal numbers

    * importer : str, None
               : Path to .js file containing custom importer

    * functions : str, None
                : Path to .js file containing custom functions

Requirements:

    * node.js
    * node-sass
      to install, run `npm install --save-dev node-sass`

"""
import os
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    conf = {
        '_replace_patterns_': ((r'\.scss$', '.css'),)
    }
    workdir = None

    def prepare(self):
        cfg = self.conf

        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)

        self.add_bool_args('omit_source_map_url', 'indented_syntax',
                'source_comments', 'source_map_contents', 'source_map_embed',
                'follow')

        self.add_int_args('indent_width', 'precision')

        self.add_path_args('source_map_root', 'importer', 'functions')

        self.add_path_list_args_multi('include_path')

        self.add_str_args('output_style', 'indent_type', 'linefeed',
                'source_map')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal("%s only need one input, got %s" % (
                    tool_name.capitalize(), repr(self.file_in)))

        if len(self.file_out) != 1:
            self.bld.fatal("%s can only have one output, got %s" % (
                    tool_name.capitalize(), repr(self.file_out)))

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} < {in_} > {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ),
        **kwargs)


def configure(conf):
    bin_path = 'node_modules/node-sass/bin/node-sass'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.abspath(bin_path)
    else:
        bin_path = conf.find_program('node-sass')[0]
    conf.end_msg(bin_path)
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



import os
import subprocess
import sys

def expand_resource(group, path):
    """Get real path of a resource."""

    bld = group.context
    # replacement pattern, {_N} will be replaced with group name of level N
    path = path.format(**group.get_patterns())
    path = os.path.expanduser(path)
    if os.path.isabs(path):
        if path.endswith(os.path.sep):
            node = bld.root.find_dir(path.lstrip('/'))
        else:
            node = bld.root.find_resource(path.lstrip('/'))
        if node:
            return node.abspath()
    else:
        if path.endswith(os.path.sep):
            node = bld.path.find_dir(path)
        else:
            node = bld.path.find_resource(path)
        if node:
            return node.abspath()
    return None


def expand_wildcard(group, path):
    """Get real path of a resource."""

    bld = group.context
    # replacement pattern, {_N} will be replaced with group name of level N
    path = path.format(**group.get_patterns())
    if '*' in path or '?' in path:
        if os.path.isabs(path):
            return [node.abspath() for node in\
                    bld.root.ant_glob(path.lstrip('/'))]
        return [node.abspath() for node in bld.path.ant_glob(path)]

    path = os.path.expanduser(path)
    if os.path.isabs(path):
        if path.endswith(os.path.sep):
            node = bld.root.find_dir(path.lstrip('/'))
        else:
            node = bld.root.find_resource(path.lstrip('/'))
        if node:
            return [node.abspath()]

    else:
        if path.endswith(os.path.sep):
            node = bld.path.find_dir(path)
        else:
            node = bld.path.find_resource(path)
        if node:
            return [node.abspath()]

    return []


def PATH(path):
    if sys.platform in ('cygwin', 'msys'):
        unix_path = subprocess.check_output(['cygpath', '-u', path])
        return unix_path.decode().rstrip()

    return path



"""
Patch apply changes to a file

Options:

    * patch_file : str, None, location of patch file (can be made using this
                   command `diff -Naur original_file current_file` > patch_file

Requirements:

    * patch

"""
import os
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        if cfg.get('patch_file') is None:
            self.bld.fatal('InvalidOptions: "patch_file" is missing')
        args.append('-i ' + os.path.realpath(cfg['patch_file']))


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output' %\
                    tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} -o {out} {in_}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program('patch')[0]



""" Render a mustache template with the given context.

Requirements:

    * pystache
      to install, run `pip install pystache`

"""
import os
from pystache import render
#-
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal("%s only need one input" % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal("%s can only have one output" %\
                    tool_name.capitalize())

        context = self.conf.get('context', {})
        os.makedirs(os.path.dirname(self.file_out[0]), exist_ok=True)
        with open(self.file_out[0], 'w', encoding='utf-8') as fout:
            with open(self.file_in[0], 'r', encoding='utf-8') as fin:
                fout.write(render(fin.read(), dict(context)))

        return 0



"""
Validate javascript files.

Options:

    * config_file      : str, None, jscs configuration file
    * esnext           : bool, False, attempts to parse esnext code
                        (currently es6)
    * esprima          : str, None, attempts to use a custom version of Eprima
    * colors           : bool, True, output with colors
    * preset           : str, None, preset config
    * verbose          : adds rule names to the error output
    * max_errors       : int, None, maximum number of errors to report
    * error_filter     : str, None, a module to filter errors
    * reporter         : str, None, error reporter, console - default, text,
                         checkstyle, junit, inline
                         also accepts relative or absolute path to custom
                         reporter

Requirements:

    * node.js
    * jscs
      to install, run `npm install --save-dev jscs`

"""

import os
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_source_grouped_': True,
    }
    name = tool_name

    def prepare(self):
        bld = self.group.context
        cfg = self.conf
        args = self.args

        # Custom configuration file
        c = cfg.get('config_file', None)
        if c:
            args.append("--config='%s'" % bld.path.find_resource(c).abspath())

        c = cfg.get('esnext', False)
        if c:
            args.append('--esnext')

        c = cfg.get('esprima', None)
        if c:
            args.append("--esprima='%s'" % bld.path.find_resource(c).abspath())

        c = cfg.get('colors', True)
        if not c:
            args.append('--no-colors')

        c = cfg.get('preset', None)
        if c:
            args.append("--preset='%s'" % c)

        c = cfg.get('verbose', False)
        if c:
            args.append('--verbose')

        c = cfg.get('max_errors', 0)
        if c > 0:
            args.append("--max-errors='%s'" % c)

        c = cfg.get('error_filter', None)
        if c:
            args.append("--error-filter='%s'" % c)

        if cfg.get('reporter', None):
            args.append('--reporter=%s' % cfg['reporter'])


    def perform(self):
        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=' '.join(self.file_in),
        ))


def configure(conf):
    bin_path = 'node_modules/jscs/bin/jscs'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('jscs')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Generate diagram.

Options:

    * work_dir : str, None
               : Change current directory before running

    * graph_attr : dict, None
                 : Set graph attribute 'name' to 'val'

    * node_attr  : dict, None
                 : Set node attribute 'name' to 'val'

    * edge_attr  : dict, None
                 : Set edge attribute 'name' to 'val'

    * output_format : str, None
                    : Set output format
                    : Possible values:
                    :   - bmp
                    :   - canon
                    :   - cmap
                    :   - cmapx
                    :   - cmapx_np
                    :   - dot
                    :   - eps
                    :   - fig
                    :   - gd
                    :   - gd2
                    :   - gif
                    :   - gtk
                    :   - gv
                    :   - ico
                    :   - imap
                    :   - imap_np
                    :   - ismap
                    :   - jpe
                    :   - jpeg
                    :   - jpg
                    :   - pdf
                    :   - pic
                    :   - plain
                    :   - plain-ext
                    :   - png
                    :   - pov
                    :   - ps
                    :   - ps2
                    :   - svg
                    :   - svgz
                    :   - tif
                    :   - tiff
                    :   - tk
                    :   - vml
                    :   - vmlz
                    :   - vrml
                    :   - wbmp
                    :   - x11
                    :   - xdot
                    :   - xdot1.2
                    :   - xdot1.4
                    :   - xlib

    * layout_engine : str, None
                    : Set layout engine
                    : Possible values:
                    :   - circo
                    :   - dot
                    :   - fdp
                    :   - neato
                    :   - nop
                    :   - nop1
                    :   - nop2
                    :   - osage
                    :   - patchwork
                    :   - sfdp
                    :   - twopi

    * library : list, None
              : Use external library

    * verbosity : int, None
                : Set level of message suppression (default=1)

    * scale : int, None
            : Scale input by n (default=72)

    * v_flip : bool, None
             : Invert y coordinate in output

    * no_layout : int, None
                : No layout mode n (default=1)

    * reduce_graph : bool, None
                   : Reduce graph

    * no_grid : bool, None
              : Don't use grid

    * old_attract : bool, None
                  : Use old attractive force

    * iteration : int, None
                : Set number of iterations to n

    * unscaled_factor : int, None
                      : Set unscaled factor to n

    * overlap_expansion_factor : int, None
                               : Set overlap expansion factor to n

    * temperature_factor : int, None
                         : Set temperature (temperature factor) to n

Requirements:

    * graphviz
      to install, for example run `apt-get install graphviz`

"""
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        arg = self.args

        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)

        c = cfg.get('graph_attr')
        if isinstance(c, dict):
            for key in c:
                arg.append('-G%s=%s' % (key, c[key]))
        elif c is not None:
            self.bld.fatal('graph_attr must be a dictionary.')

        c = cfg.get('node_attr')
        if isinstance(c, dict):
            for key in c:
                arg.append('-N%s=%s' % (key, c[key]))
        elif c is not None:
            self.bld.fatal('node_attr must be a dictionary.')

        c = cfg.get('edge_attr')
        if isinstance(c, dict):
            for key in c:
                arg.append('-E%s=%s' % (key, c[key]))
        elif c is not None:
            self.bld.fatal('edge_attr must be a dictionary.')

        c = cfg.get('output_format')
        if c:
            arg.append('-T' + c)

        c = cfg.get('layout_engine')
        if c:
            arg.append('-K' + c)

        c = cfg.get('library', [])
        for lib in c:
            arg.append('-l' + lib)

        c = cfg.get('verbosity')
        if c is not None:
            arg.append('-q' + c)

        c = cfg.get('scale')
        if c:
            arg.append('-s' + c)

        c = cfg.get('v_flip')
        if c:
            arg.append('-y')

        c = cfg.get('no_layout')
        if c:
            arg.append('-n' + c)

        c = cfg.get('reduce_graph')
        if c:
            arg.append('-x')

        c = cfg.get('no_grid')
        if c:
            arg.append('-Lg')

        c = cfg.get('old_attract')
        if c:
            arg.append('-LO')

        c = cfg.get('iteration')
        if c:
            arg.append('-Ln' + c)

        c = cfg.get('unscaled_factor')
        if c:
            arg.append('-LU' + c)

        c = cfg.get('overlap_expansion_factor')
        if c:
            arg.append('-LC' + c)

        c = cfg.get('temperature_factor')
        if c:
            arg.append('-LT' + c)


    def perform(self):
        if not self.file_in:
            self.bld.fatal('%s needs input' % tool_name)
        if len(self.file_out) != 1:
            self.bld.fatal('%s produces one output' % tool_name)

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command([executable, '-o' + self.file_out[0]] +\
                self.args + self.file_in, **kwargs)


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program(tool_name)[0]



import os
from pybuildtool.misc.resource import get_source_files
from subprocess import call
import sys
from time import sleep
import yaml

from .file_observer import FileObserver

class Application(object):

    bld = None
    observer = None

    rebuild = True
    reload = True
    running = True

    config_file = None
    sysargs = None

    def __init__(self, bld):
        self.bld = bld
        self.watchers = {}
        self.config_file = os.path.realpath(os.path.join(bld.path.abspath(),
                'build.yml'))

        self.sysargs = sys.argv[:]
        # Something from [fireh_runner](https://github.com/dozymoe/fireh_runner],
        # when running under PYTHONUSERBASE, let's use approriate python binary.
        python_bin = os.environ.get('PYTHON_BIN')
        if python_bin:
            self.sysargs.insert(0, python_bin)

        for ii in range(1, len(self.sysargs)):
            if self.sysargs[ii] == 'watch' or\
                    self.sysargs[ii].startswith('watch_'):

                self.sysargs[ii] = self.sysargs[ii].replace('watch',
                        'build', 1)
                break

        self.observer = FileObserver(self)


    def run(self):
        while self.running:
            if self.reload:
                self.do_reload()

            if self.rebuild:
                self.rebuild = False
                # On windows we'd be using waf.bat
                use_shell = os.name == 'nt'
                call(self.sysargs, shell=use_shell)

            count = 0
            while count < 10 and self.running:
                count += 1
                sleep(1)

        print('Closing files observers..')
        self.observer.close()


    def do_reload(self):
        self.reload = False
        self.observer.close()

        with open(self.config_file, 'r') as f:
            config = yaml.load(f, Loader=yaml.SafeLoader)

        self.observer.open(list(os.path.realpath(f) for f in\
                get_source_files(config, self.bld)))

        self.rebuild = True


    def stop(self, *args):
        self.running = False



"""
scp is secure copy (remote file copy program)

Options:

    * username     : str, None, use the user for authentication
    * identify_file: str, None, use the ssh private key file for authentication
    * host         : str, None, host name, URL or bookmark name
    * port         : int, None, use the port for connection

Requirements:

    * scp
      to install, for example run `apt-get install openssh`

"""
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_source_grouped_': True,
    }
    hoststr = None
    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        # identity file
        c = cfg.get('identity_file', None)
        if c:
            args.append('-i %s' % expand_resource(self.group, c))

        # port
        c = cfg.get('port', None)
        if c:
            args.append('-P %s' % c)

        # host
        h = cfg.get('host', None)
        if h:
            u = cfg.get('username', None)
            if u:
                self.hoststr = '%s@%s' % (u, h)
            else:
                self.hoststr = h
        else:
            self.bld.fatal('configuration "%s" is required for %s' % (
                    'host', tool_name))


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            "{exe} {arg} {in_} {hst}:{out}".format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
            hst=self.hoststr,
        ))


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program('scp')[0]



"""
Base class for pybuildtools tools.

Options:

    * _source_excluded_ : list, None
                        : Pretend source files (*_in values) don't exist.

    * _source_basedir_ : str, None
                       : Create files in output dir, relative path to source
                       : base directory.

    * _source_grouped_ : bool, None
                       : Don't create separate tasks for every input files, have
                       : them as input files of a single task.
                       : Actually I'm not so sure what this does, something like
                       : have them all as arguments to shell command?

    * _noop_retcodes_ : list, None
                      : If Task.perform() returns these, pretend nothing
                      : happened.

    * _success_retcodes_ : list, None
                         : If Task.perform() returns these, pretend as if it
                         : returns 0 or a success.

    * _replace_patterns_ : list, None
                         : If the output is a directory, you can rename the
                         : output files based on the source files.
                         : This is a list of list.
                         : The list elements consist of two items: python regex
                         : and replacement.

    * _no_io_ : bool, False
              : This task doesn't need inputs or outputs.
              : Only works if written in build.yml.

"""
import os
from copy import deepcopy
from time import time
from uuid import uuid4
import stringcase
from waflib.Task import Task as BaseTask # pylint:disable=import-error

from ..misc.collections_utils import make_list
from ..misc.path import expand_resource

class Task(BaseTask):

    args = None
    args_case = 'spinal'
    conf = None
    group = None
    file_in = None
    file_out = None
    name = None
    token_in = None
    token_out = None

    _id = None

    def __init__(self, group, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._id = uuid4().hex

        # Task's configuration can be declared higher in the build tree,
        # but it needs to be prefixed with its tool-name.
        # Tool-name however can only be defined by the tool's module by
        # observing predefined `__name__` variable, which value is the name
        # of the tool's module.
        if config:
            my_config = deepcopy(config)
            if self.name:
                name = self.name + '_'
                for key in config.keys():
                    if not key.startswith(name):
                        continue
                    task_conf = key[len(name):]
                    if task_conf in config:
                        continue
                    my_config[task_conf] = config[key]
        else:
            my_config = {}
        self.args = []
        self.conf = my_config
        self.group = group
        self.file_in = []
        self.file_out = []
        self.token_in = []
        self.token_out = []


    def prepare(self):
        pass


    def prepare_args(self): # pylint:disable=no-self-use
        return []


    def prepare_shadow_jutsu(self):
        source_exclude = []
        for f in make_list(self.conf.get('_source_excluded_')):
            nodes = expand_resource(self.group, f)
            source_exclude += make_list(nodes)

        task_uid = self._id

        for node in self.inputs:
            path = node.abspath()
            if node.parent.name == '.tokens':
                self.token_in.append(path)
            elif getattr(node, 'is_virtual_in_' + task_uid, False):
                pass
            elif path in source_exclude:
                pass
            else:
                self.file_in.append(path)

        for node in self.outputs:
            path = node.abspath()
            if node.parent.name == '.tokens':
                self.token_out.append(path)
            elif getattr(node, 'is_virtual_out_' + task_uid, False):
                pass
            else:
                self.file_out.append(path)


    def finalize_shadow_jutsu(self, create_only=False):
        for filename in self.token_out:
            if create_only and os.path.exists(filename):
                continue
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError:
                pass
            with open(filename, 'w', encoding='utf-8') as f:
                f.write(str(time()))


    def run(self):
        self.prepare_shadow_jutsu()
        self.prepare()
        ret = self.perform()

        create_only = False
        if ret in make_list(self.conf.get('_noop_retcodes_')):
            create_only = True
            ret = None
        elif ret in make_list(self.conf.get('_success_retcodes_')):
            ret = None
        if not ret:
            self.finalize_shadow_jutsu(create_only)
        return ret


    @staticmethod
    def is_production():
        return os.environ.get('PROJECT_VARIANT_IS_PRODUCTION') == '1'


    def stringcase_arg(self, option):
        return getattr(stringcase, self.args_case + 'case')(option)


    def _add_arg(self, option, value, sep):
        if sep == ' ':
            self.args.append(option)
            self.args.append(value)
        else:
            self.args.append(option + sep + value)


    def add_bool_args(self, *options):
        for option in options:
            value = self.conf.get(option)
            if not value:
                continue

            option = '--' + self.stringcase_arg(option)
            self.args.append(option)


    def add_dict_args(self, *options, **kwargs):
        opt_val_sep = kwargs.get('opt_val_sep', '=')
        key_val_sep = kwargs.get('key_val_sep', '=')

        for option in options:
            if option not in self.conf:
                continue

            for key, value in self.conf[option].items():
                value = value.format(**self.group.get_patterns())
                item = key + key_val_sep + value
                self._add_arg('--' + option, item, opt_val_sep)


    def add_int_args(self, *options, **kwargs):
        opt_val_sep = kwargs.get('opt_val_sep', '=')

        for option in options:
            try:
                value = int(self.conf.get(option))
            except (TypeError, ValueError):
                continue

            option = '--' + self.stringcase_arg(option)
            self._add_arg(option, str(value), opt_val_sep)


    def add_list_args_join(self, separator, *options, **kwargs):
        opt_val_sep = kwargs.get('opt_val_sep', '=')

        for option in options:
            values = make_list(self.conf.get(option))
            if not values:
                continue

            option = '--' + self.stringcase_arg(option)
            value = separator.join(x.format(**self.group.get_patterns())\
                    for x in values)

            self._add_arg(option, value, opt_val_sep)


    def add_list_args_multi(self, *options, **kwargs):
        opt_val_sep = kwargs.get('opt_val_sep', '=')

        for option in options:
            values = make_list(self.conf.get(option))
            if not values:
                continue

            option = '--' + self.stringcase_arg(option)
            for value in values:
                value = value.format(**self.group.get_patterns())

                self._add_arg(option, value, opt_val_sep)


    def add_path_args(self, *options, **kwargs):
        opt_val_sep = kwargs.get('opt_val_sep', '=')

        for option in options:
            value = self.conf.get(option)
            if value is None:
                continue

            option = '--' + self.stringcase_arg(option)
            value = expand_resource(self.group, value)
            self._add_arg(option, value, opt_val_sep)


    def add_path_list_args_join(self, separator, *options, **kwargs):
        opt_val_sep = kwargs.get('opt_val_sep', '=')

        for option in options:
            values = make_list(self.conf.get(option))
            if not values:
                continue

            option = '--' + self.stringcase_arg(option)
            value = separator.join(expand_resource(self.group, x)\
                    for x in values)

            self._add_arg(option, value, opt_val_sep)


    def add_path_list_args_multi(self, *options, **kwargs):
        opt_val_sep = kwargs.get('opt_val_sep', '=')

        for option in options:
            values = make_list(self.conf.get(option))
            if not values:
                continue

            option = '--' + self.stringcase_arg(option)
            for value in values:
                value = expand_resource(self.group, value)
                self._add_arg(option, value, opt_val_sep)


    def add_str_args(self, *options, **kwargs):
        opt_val_sep = kwargs.get('opt_val_sep', '=')

        for option in options:
            value = self.conf.get(option)
            if value is None:
                continue

            option = '--' + self.stringcase_arg(option)
            value = value.format(**self.group.get_patterns())
            self._add_arg(option, value, opt_val_sep)



"""
create and edit fonts in many formats: OpenType, TrueType, AAT, PostScript,
Multiple Master, CID-Keyed, SVG and various bitmap formats

Requirements:

    * fontforge
      to install, run `apt-get install fontforge` (for example)

"""
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_replace_patterns_': ((r'\.woff$', '.svg'), (r'\.woff2$', '.svg'))
    }
    name = tool_name

    def prepare(self):
        args = self.args
        args.append('-lang ff')
        args.append("-c 'Open($1); Generate($2)'")


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    bin_path = conf.find_program('fontforge')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
lftp is sophisticated file transfer progam

Options:

    * username: str, None, use the user for authentication
    * password: str, None, use the password for authentication
    * host    : str, None, host name, URL or bookmark name
    * port    : int, None, use the port for connection

Requirements:

    * lftp
      to install, for example run `apt-get install lftp`

"""

from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_source_grouped_': True,
    }
    hoststr = None
    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        # user and password
        u = cfg.get('username', None)
        if u:
            p = cfg.get('password', None)
            if p:
                authstr = '%s,%s' % (u, p)
            else:
                authstr = u
            args.append('-u %s' % authstr)

        # port
        c = cfg.get('port', None)
        if c:
            args.append('-p %s' % c)

        # host
        c = cfg.get('host', None)
        if c:
            self.hoststr = c
        else:
            self.bld.fatal('configuration "%s" is required for %s' % (
                    'host', tool_name))


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            "{exe} {arg} -e 'put {in_} -o {out}' {hst}".format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
            hst=self.hoststr,
        ))


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program('lftp')[0]



r'''
Restores NuGet packages.

If a solution is specified, this command restores NuGet packages that are
installed in the solution and in projects contained in the solution. Otherwise,
the command restores packages listed in the specified packages.config file,
Microsoft Build project, or project.json file.

Options:
    * require_consent : bool, None
                      : Checks if package restore consent is granted before
                      : installing a package.

    * project_to_project_timeout : int, None
                                 : Timeout in seconds for resolving project to
                                 : project references.

    * packages_directory : str, None
                         : (OutputDirectory) Specifies the packages folder.

    * solution_directory : str, None
                         : Specifies the solution directory. Not valid when
                         : restoring packages for a solution.

    * msbuild_version : int, None
                      : Specifies the version of MSBuild to be used with this
                      : command.
                      : Supported values are 4, 12, 14.
                      : By default the MSBuild in your path is picked, otherwise
                      : it defaults to the highest installed version of MSBuild.

    * msbuild_path : str, None
                   : Specifies the path of MSBuild to be used with this command.
                   : This command will takes precedence over MSbuildVersion,
                   : nuget will always pick MSbuild from this specified path.

    * recursive : bool, None
                : Restore all referenced projects for UWP and NETCore projects.
                : This does not include packages.config projects.

    * source : list, None
             : A list of packages sources to use for this command.

    * fallback_source : list, None
                      : A list of package sources to use as fallbacks for this
                      : command.

    * no_cache : bool, None
               : Disable using the machine cache as the first package source.

    * direct_download : bool, None
                      : Download directly without populating any caches with
                      : metadata or binaries.

    * disable_parallel_processing : bool, None
                                  : Disable parallel processing of packages for
                                  : this command.

    * package_save_mode : str, None
                        : Specifies types of files to save after package
                        : installation: nuspec, nupkg, nuspec;nupkg.

    * verbosity : str, None
                : Display this amount of details in the output: normal, quiet,
                : detailed.

    * non_interactive : bool, None
                      : Do not prompt for user input or confirmations.

    * config_file : str, None
                  : The NuGet configuration file. If not specified,
                  : file %AppData%\NuGet\NuGet.config is used as configuration
                  : file.

    * force_english_output : bool, None
                           : Forces the application to run using an invariant,
                           : English-based culture.

Requirements:

    * nuget cli

'''
from pybuildtool import BaseTask, expand_resource, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        cfg = self.conf
        arg = self.args

        c = cfg.get('require_consent')
        if c:
            arg.append('-RequireConsent')

        c = cfg.get('project_to_project_timeout')
        if c:
            arg.append('-Project2ProjectTimeOut')
            arg.append('%i' % int(c))

        c = cfg.get('packages_directory')
        if c:
            arg.append('-PackagesDirectory')
            arg.append(c.format(**self.group.get_patterns()))

        c = cfg.get('solution_directory')
        if c:
            arg.append('-SolutionDirectory')
            arg.append(c.format(**self.group.get_patterns()))

        c = cfg.get('msbuild_version')
        if c:
            arg.append('-MSBuildVersion')
            arg.append('%i' % int(c))

        c = cfg.get('msbuild_path')
        if c:
            path = expand_resource(self.group, c)
            if path is None:
                self.bld.fatal('msbuild_path not found: ' + c)
            arg.append('-MSBuildPath')
            arg.append(path)

        c = cfg.get('recursive')
        if c:
            arg.append('-Recursive')

        sources = make_list(cfg.get('source'))
        for source in sources:
            path = expand_resource(self.group, source)
            if path is None:
                self.bld.fatal('source not found: ' + c)
            arg.append('-Source')
            arg.append(path)

        sources = make_list(cfg.get('fallback_source'))
        for source in sources:
            path = expand_resource(self.group, source)
            if path is None:
                self.bld.fatal('fallback_source not found: ' + c)
            arg.append('-FallbackSource')
            arg.append(path)

        c = cfg.get('no_cache')
        if c:
            arg.append('-NoCache')

        c = cfg.get('direct_download')
        if c:
            arg.append('-DirectDownload')

        c = cfg.get('disable_parallel_processing')
        if c:
            arg.append('-DisableParallelProcessing')

        c = cfg.get('package_save_mode')
        if c:
            arg.append('-PackageSaveMode')
            arg.append(c)

        c = cfg.get('verbosity')
        if c:
            arg.append('-Verbosity')
            arg.append(c)

        c = cfg.get('non_interactive')
        if c:
            arg.append('-NonInteractive')

        c = cfg.get('config_file')
        if c:
            path = expand_resource(self.group, c)
            if path is None:
                self.bld.fatal('config_file not found: ' + c)
            arg.append('-ConfigFile')
            arg.append(path)

        c = cfg.get('force_english_output')
        if c:
            arg.append('-ForceEnglishOutput')


    def perform(self):
        executable = self.env['%s_BIN' % tool_name.upper()]
        cmds = [executable, 'restore'] + self.args + self.file_in
        return self.exec_command(cmds)


def configure(conf):
    bin_path = conf.find_program('nuget.exe')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Convert google font's METADATA.json into css @font-face.

Options:

    * font_dir_url : str, ''
                     The directory url on where the file was going to be hosted.
                     The value should ends with '/'.
    * font_types   : list, ['ttf']
                     Print the stated font types.
    * font_svg_id  : str, None
                     Default is the font's PostScript name.

Requirements:

    * protobuf
      to install, run `pip install protobuf`

"""
import json
from pybuildtool import BaseTask, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    dir_url = None
    print_ttf = None
    print_eot = None
    print_woff = None
    print_svg = None
    svg_id = None

    def prepare(self):
        cfg = self.conf

        types = make_list(cfg.get('font_types', ['ttf']))
        self.print_ttf = 'ttf' in types
        self.print_eot = 'eot' in types
        self.print_woff = 'woff' in types
        self.print_svg = 'svg' in types

        self.dir_url = cfg.get('font_dir_url', '').format(**self.group.\
                get_patterns())

        self.svg_id = cfg.get('font_svg_id')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only needs one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only needs one output' % tool_name.capitalize())

        filename = self.file_in[0]
        if filename.endswith('.json'):
            with open(self.file_in[0], 'r', encoding='utf-8') as f:
                fonts = json.load(f)['fonts']
        elif filename.endswith('.pb'):
            fonts = self._get_fonts_from_protobuf(filename)
        else:
            self.bld.fatal('Unknown file format: ' + filename)

        with open(self.file_out[0], 'w', encoding='utf-8') as f:
            for font in fonts:
                f.write('@font-face {\n')
                f.write('    font-family: "%s";\n' % font['name'])
                f.write('    font-style: %s;\n' % font['style'])
                f.write('    font-weight: %i;\n' % font['weight'])
                f.write('    src:\n')
                f.write('        local("%s")' % font['fullName'])

                arg = {'dir': self.dir_url, 'name': font['postScriptName']}

                if self.print_eot:
                    f.write((',\n        url({dir}{name}.eot?#iefix) ' +\
                            'format("embedded-opentype")').format(**arg))

                if self.print_woff:
                    f.write((',\n        url({dir}{name}.woff) ' +\
                            'format("woff")').format(**arg))

                if self.print_ttf:
                    f.write((',\n        url({dir}{name}.ttf) ' +\
                            'format("truetype")').format(**arg))

                if self.print_svg:
                    arg['id'] = self.svg_id or font['postScriptName']
                    f.write((',\n        url({dir}{name}.svg#{id}) ' +\
                            'format("svg")').format(**arg))

                f.write(';\n}\n')


    @staticmethod
    def _get_fonts_from_protobuf(filename):
        from google.protobuf import text_format # pylint:disable=import-error,no-name-in-module,import-outside-toplevel
        from pybuildtool.vendor.fonts_public_pb2 import FamilyProto # pylint:disable=import-outside-toplevel

        message = FamilyProto()
        with open(filename, encoding='utf-8') as f:
            text_format.Merge(f.read(), message)

        for font in message.fonts:
            yield {t[0].json_name: t[1] for t in font.ListFields()}



'''
CSS-Live-Reloader notifier

see: https://github.com/sergiosgc/CSS-Live-Reloader

Requirements:

    * ws4py
      to install, run `pip install ws4py`

'''
from threading import Thread
from wsgiref.simple_server import make_server
from ws4py.websocket import WebSocket
from ws4py.server.wsgirefserver import WSGIServer, WebSocketWSGIRequestHandler
from ws4py.server.wsgiutils import WebSocketWSGIApplication

clients = set()

class BrowserReloadProtocol(WebSocket):

    def closed(self, code, reason):
        global clients
        clients.discard(self)


    def opened(self):
        global clients
        clients.add(self)
        print('css-reload client', self.peer_address)


    def received_message(self, message):
        print(message)


class BrowserReloadNotifier(object):

    host = None
    server = None
    server_thread = None

    def __init__(self, host):
        self.host = host


    def start(self):
        if self.server:
            return
        self.server = make_server(self.host, 3000, server_class=WSGIServer,
                handler_class=WebSocketWSGIRequestHandler,
                app=WebSocketWSGIApplication(handler_cls=BrowserReloadProtocol))
        self.server.initialize_websockets_manager()
        self.server_thread = Thread(target=self.server.serve_forever)
        self.server_thread.start()


    def stop(self):
        if self.server is None:
            return
        for client in clients:
            client.close()
        self.server.shutdown()
        self.server_thread.join()


    def trigger(self):
        for client in clients:
            client.send('notification\n')



"""
Generate documentation.

Options:

    * work_dir      : str, None, change current directory
    * config_file   : str, None, doxygen configuration file

Requirements:

    * doxygen
      to install, for example run `apt-get install doxygen`

"""
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        # Change current directory, before running pylint, helps module imports
        c = cfg.get('work_dir', None)
        if c:
            self.workdir = expand_resource(self.group, c)
            if self.workdir is None:
                self.bld.fatal(cfg['work_dir'] + ' not found.')

        # Specify a configuration file
        c = cfg.get('config_file', None)
        if c:
            args.append(expand_resource(self.group, c))


    def perform(self):
        if self.file_in:
            self.bld.fatal('%s takes no input' % tool_name.capitalize())
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg}'.format(
                exe=executable,
                arg=' '.join(self.args),
            ),
            **kwargs)


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program(tool_name)[0]



'''
Run ansible module

Options:

    * module        : str, None, ansible module name to run
    * args          : dict, None, arguments for the module
    * free_form     : str, None, modules like `shell` and `command`, besides
                      args, also take in free_form arg
                      we facilitate this to provide direct access to
                      `modules_args`, our args option will actually be send as
                      `complex_args`
    * with_items    : str, None, dot separated dictionary keys for `context`
    * hosts         : list, None, host names from ANSIBLE_HOSTS file
    * hosts_pattern : str, '*', limit target hosts to a pattern
    * forks         : int, None, maximum number of processes to spawn
    * connect_timeout : int, None, ssh timeout
    * context_python: str, None, python scripts containing context for template
                      if it has __init__.py in the same directory it will be
                      added to sys.path
                      it must have variable type dict named: `export`
    * context_yaml  : str, None, yaml file containing context for template
    * context       : any, {}, context to be used in template

Requirements:

    * ansible
      to install, run `pip install ansible`

'''
import os
import sys
import six
from waflib import Logs # pylint:disable=import-error
from yaml import safe_load as yaml_load
#-
from pybuildtool import BaseTask, expand_resource, make_list
from pybuildtool.misc.python import load_module_from_filename

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    context = None
    items = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context = {}
        self.items = []


    def prepare(self):
        c = self.conf.get('module')
        if c is None:
            self.bld.fatal('"module" option in required')
        elif os.path.sep in c:
            c = expand_resource(self.bld, c)
            dirname, basename = os.path.split(c)
            modulename, extname = os.path.splitext(basename)
            assert extname == '.py'
            self.args.append(('module_path', dirname))
            self.args.append(('module_name', modulename))
        else:
            self.args.append(('module_name', c))

        c = self.conf.get('args')
        if c:
            assert isinstance(c, dict)
            self.args.append(('complex_args', c))

        c = self.conf.get('free_form')
        if c:
            self.args.append(('module_args', c))

        c = self.conf.get('forks')
        if c:
            try:
                c = int(c)
            except ValueError:
                self.bld.fatal('"forks" option must be numeral')
            self.args.append(('forks', c))

        c = self.conf.get('connect_timeout')
        if c:
            try:
                c = int(c)
            except ValueError:
                self.bld.fatal('"timeout" option must be numeral')
            self.args.append(('timeout', c))

        c = self.conf.get('hosts_pattern', '*')
        self.args.append(('pattern', c))

        c = make_list(self.conf.get('hosts'))
        if c:
            self.args.append(('run_hosts', c))

        # do not run async
        self.args.append(('background', 0))

        ## context

        # Yaml context
        c = self.conf.get('context_yaml')
        if c:
            yaml_file = expand_resource(self.group, c)
            if yaml_file is None:
                self.bld.fatal('"context_yaml" for %s has invalid value' %\
                        tool_name.capitalize())
            with open(yaml_file, 'r', encoding='utf-8') as f:
                self.context.update(yaml_load(f))

        # Python context
        c = self.conf.get('context_python')
        if c:
            python_file = expand_resource(self.group, c)
            if python_file is None:
                self.bld.fatal('"context_python" for %s has invalid value' %\
                        tool_name.capitalize())
            dirname, filename = os.path.split(python_file)
            filebase, _ = os.path.splitext(filename)
            if os.path.exists(os.path.join(dirname, '__init__.py')):
                sys.path.append(dirname)
                mod = __import__(filebase)
            else:
                mod = load_module_from_filename(python_file, filebase)
            python_export = mod.export
            self.context.update(python_export)

        self.context.update(self.conf.get('context', {}))
        self.args.append(('play_vars', self.context))

        # with_items
        c = self.conf.get('with_items')
        if isinstance(c, list):
            wi = list(c)
        elif c:
            c = c.split('.')
            wi_ctx = self.context
            for wi_idx in c:
                wi = wi_ctx[wi_idx]
            wi = make_list(wi, nodict=True)
        else:
            wi = []

        for wi_item in wi:
            if isinstance(wi_item, dict):
                self.items.append(wi_item)
            else:
                self.items.append({'item': wi_item})


    def perform(self):
        import ansible.runner # pylint:disable=import-error,import-outside-toplevel

        def log_error(data):
            Logs.error('Got "{msg}" from {host}'.format(
                tool=tool_name, **data))

        def extract_errors(result):
            print(result)
            changed = False
            success = True
            for host in result['contacted']:
                if result['contacted'][host].get('failed'):
                    log_error({
                        'host': host,
                        'msg': result['contacted'][host]['msg'],
                    })
                    success = False
                if result['contacted'][host].get('changed'):
                    changed = True
            return success, changed

        success = True
        changed = False
        if self.items:
            args = dict(self.args)
            kwargs = args['complex_args']
            for item in self.items:
                item_kwargs = {}
                for field in kwargs:
                    if isinstance(kwargs[field], six.string_types):
                        item_kwargs[field] = kwargs[field] % item
                    else:
                        item_kwargs[field] = kwargs[field]
                args['complex_args'] = item_kwargs
                result = ansible.runner.Runner(**args).run()
                was_success, was_changed = extract_errors(result)
                success = success and was_success
                changed = changed or was_changed
        else:
            result = ansible.runner.Runner(**dict(self.args)).run()
            was_success, was_changed = extract_errors(result)
            success = success and was_success
            changed = changed or was_changed

        #if changed:
        #    self.finalize_shadow_jutsu(use_file_out=True)
        #elif success:
        #    for fo in self.file_out:
        #        if not os.path.exists(fo):
        #            self.finalize_shadow_jutsu(use_file_out=True)
        #            break

        if success:
            return 0
        return 1


def configure(conf):
    conf.start_msg("Checking for python module '%s'" % tool_name)
    try:
        import ansible # pylint: disable=unused-import,import-outside-toplevel
    except ImportError:
        conf.end_msg('not found', color='YELLOW')



"""
Splint is a tool for statically checking C programs for security vulnerabilites
and coding mistakes. With minimal effort, Splint can be used as a better lint.
If additional effort is invested adding annotations to programs, Splint can
perform stronger checking than can be done by any standard lint.

Options:

    * work_dir : str, None, Change current directory
    * flags    : list, [], enable or disable checks

Requirements:

    * splint
      to install, for example run `apt-get install splint`

"""
from pybuildtool import BaseTask, expand_resource, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        # Change current directory, before running pylint, helps module imports
        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)
            if self.workdir is None:
                self.bld.fatal(cfg['work_dir'] + ' not found.')

        # Flags
        c = make_list(cfg.get('flags'))
        if not c:
            # see:
            # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=498809;msg=10
            c = [
                '-D__gnuc_va_list=va_list',
                '-warnposix',
                '+forcehints',
                '-formatcode',
                '-compdestroy',
            ]

        for flag in c:
            args.append(flag)


    def perform(self):
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
                exe=executable,
                arg=' '.join(self.args),
                in_=' '.join(self.file_in),
            ),
            **kwargs)


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program(tool_name)[0]



""" Compiles .proto into .py using protobuf.

Protocol buffers are Google's language-neutral, platform-neutral, extensible
mechanism for serializing structured data – think XML, but smaller, faster, and
simpler. You define how you want your data to be structured once, then you can
use special generated source code to easily write and read your structured data
to and from a variety of data streams and using a variety of languages.

Options:

    * proto_path : str, []
                 : Specify the directory in which to search for imports. May be
                 : specified multiple times; directories will be searched in
                 : order. If not given, the current working directory is used.
                 : If not found in any of these directories, the
                 : --descriptor_set_in descriptors will be checked for required
                 : proto file.

    * encode : str, None
             : Read a text-format message of the given type from standard input
             : and write it in binary to standard output. The message type must
             : be defined in PROTO_FILES or their imports.

    * decode : str, None
             : Read a binary message of the given type from standard input and
             : write it in the text format to standard output. The message type
             : must be defined in PROTO_FILES or their imports.

    * decode_raw : bool, None
                 : Read an arbitrary protocol message from standard input and
                 : write the raw tag/value pairs in text format to standard
                 : output. No PROTO_FILES should be given when using this flag.

    * protobuf_in : str, None
                  : Absolute path to the protobuf file from which input of
                  : encoding/decoding operation will be read. If omitted, input
                  : will be read from standard input.

    * protobuf_out : str, None
                   : Absolute path to the protobuf file to which output of
                   : encoding/decoding operation will be written. If omitted,
                   : output will be written to standard output.

    * descriptor_set_in : str, None
                        : Specifies a delimited list of FILES each containing
                        : a FileDescriptorSet (a protocol buffer defined in
                        : descriptor.proto).
                        : The FileDescriptor for each of the PROTO_FILES
                        : provided will be loaded from these FileDescriptorSets.
                        : If a FileDescriptor appears multiple times, the first
                        : occurence will be used.

    * descriptor_set_out : str, None
                         : Writes a FileDescriptorSet (a protocol buffer defined
                         : in descriptor.proto) containing all of the input
                         : files to FILE.

    * include_imports : bool, None
                      : When using --descriptor_set_out, also include all
                      : dependencies of the input files in the set, so that the
                      : set is self-contained.

    * include_source_info : bool, None
                          : When using --descriptor_set_out, do not strip
                          : SourceCodeInfo from the FileDescriptorProto.
                          : This results in vastly larger descriptors that
                          : include information about the original location of
                          : each decl in the source file as well as surrounding
                          : comments.

    * dependency_out : str, None
                     : Write a dependency output file in the format expected by
                     : make. This writes the transitive set of input file paths
                     : to FILE.

    * error_format : str, None
                   : Set the format in which to print errors.
                   : FORMAT may be 'gcc' (the default) or 'msvs' (Microsoft
                   : Visual Studio format).

    * print_free_field_numbers : bool, None
                               : Print the free field numbers of the messages
                               : defined in the given proto files. Group share
                               : the same field number space with the parent
                               : message. Extension ranges are counted as
                               : occupied fields numbers.

    * plugin : dict, {}
             : Specifies a plugin executable to use.
             : Normally, protoc searches the PATH for plugins, but you may
             : specify additional executables not in the path using this flag.
             : Additionally, EXECUTABLE may be of the form NAME=PATH, in which
             : case the given plugin name is mapped to the given executable even
             : if the executable's own name differs.

    * cpp_out : str, None
              : Generate C++ header and source

    * grpc_cpp_out : str, None
              : Generate GRPC C++ header and source

    * csharp_out : str, None
                 : Generate C# source file

    * grpc_csharp_out : str, None
                 : Generate GRPC C# source file

    * java_out : str, None
               : Generate Java source file

    * grpc_java_out : str, None
               : Generate GRPC Java source file

    * js_out : str, None
             : Generate Javascript source file

    * grpc_js_out : str, None
             : Generate GRPC Javascript source file

    * objc_out : str, None
               : Generate Objective C header and source

    * grpc_objc_out : str, None
               : Generate GRPC Objective C header and source

    * php_out : str, None
              : Generate PHP source file

    * grpc_php_out : str, None
              : Generate GRPC PHP source file

    * python_out : str, None
                 : Generate Python source file

    * grpc_python_out : str, None
                      : Generate GRPC Python source file

    * ruby_out : str, None
               : Generate Ruby source file

    * grpc_ruby_out : str, None
               : Generate GRPC Ruby source file

Requirements:

    * protoc
      to install, run `apt install protobuf-compiler`

    * grpcio-tools
      to install, run `pip install grcpio-tools`
      you might need to specify the version

"""
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    args_case = 'snake'
    name = tool_name

    def prepare(self):
        import pkg_resources # pylint:disable=import-outside-toplevel

        cfg = self.conf

        self.add_path_list_args_multi('proto_path')

        self.add_str_args('encode', 'decode', 'error_format')

        self.add_bool_args('decode_raw', 'include_imports',
                'include_source_info', 'print_free_field_numbers')

        self.add_dict_args('plugin')

        self.add_path_args('protobuf_in', 'protobuf_out',
                'descriptor_set_in', 'descriptor_set_out', 'dependency_out',
                'cpp_out', 'grpc_cpp_out',
                'csharp_out', 'grpc_csharp_out',
                'java_out', 'grpc_java_out',
                'objc_out', 'grpc_objc_out',
                'php_out', 'grpc_php_out',
                'python_out', 'grpc_python_out',
                'ruby_out', 'grpc_ruby_out')

        c = cfg.get('js_out', None)
        if c:
            if c.startswith('library='):
                c = c[:c.index('=') + 1] + expand_resource(self.group,
                        c[c.index('=') + 1:c.index(',')]) + c[c.index(','):]
            elif c.startswith('import_style='):
                c = c[:c.index(':') + 1] + expand_resource(self.group,
                        c[c.index(':') + 1:])
            else:
                c = expand_resource(self.group, c)
            self.args.append('--js_out=' + c)

        proto_include = pkg_resources.resource_filename('grpc_tools', '_proto')
        self.args.append('-I{}'.format(proto_include))


    def perform(self):
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg}'.format(
            exe=executable,
            arg=' '.join(self.args + self.file_in),
        ))


def configure(conf):
    conf.start_msg("Checking for program '%s'" % tool_name)
    bin_path = conf.find_program('protoc')[0]
    conf.end_msg(bin_path)
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
gzip compress files.

Requirements:

    * gzip

"""
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_replace_patterns_': ((r'$', '.gz'),),
    }
    name = tool_name

    def prepare(self):
        self.args = ['--stdout', '--best']


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output' %\
                    tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} > {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program('gzip')[0]



"""
Generate documentation.

Options:

    * work_dir      : str, None, change current directory
    * build_dir     : str, None, Analysis output directory. Useful for various
                      data. Some possible usages are; whole program analysis,
                      incremental analysis, distributed analysis.
    * check_library : bool, None, Show information messages when library files
                      have incomplete info.
    * config_exclude: list, [], Path (prefix) to be excluded from
                      configuration checking. Preprocessor configurations
                      defined in headers (but not sources) matching the prefix
                      will not be considered for evaluation.

    * config_exclude_file: str, None, A file that contains a list of
                           config-excludes.

    * dump          : bool, None, Dump xml data for each translation unit. The
                      dump files have the extension .dump and contain ast,
                      tokenlist, symboldatabase, valueflow.
    * define        : list, [], Define preprocessor symbol.
    * undefine      : list, [], Undefine processor symbol. Explicitly hide
                      certain #ifdef code paths from checking.
    * enable_check  : list, [], Enable additional checks. The available ids are:
                      * all
                        Enable all checks. It is recommended to only use when
                        the whole program is scanned, because this enables
                        unusedFunction.
                      * warning
                        Enable warning messages.
                      * style
                        Enable all encoding style checks. All messages with the
                        severities 'style', 'performance' and 'portability' are
                        enabled.
                      * performance
                        Enable performance messages.
                      * portability
                        Enable portability messages.
                      * information
                        Enable information messages.
                      * unusedFunction
                        Check for unused functions. It is recommend to only
                        enable this when the whole program is scanned.
                      * missingInclude
                        Warn if there are missing includes. For detailed
                        information, use '--check-config'.
    * ignore_files  : list, [], Used when certain messages should be displayed
                      but should not cause a non-zero exitcode.
    * include_path  : list, [], Give path to search for include files. First
                      given path is search for contained header files first. If
                      paths are relative to source files, this is not needed.
    * include       : list, [], Force inclusion of a file before the checked
                      file. Can be used for example when checking the Linux
                      kernel, where autoconf.h needs to be included for every
                      file compiled. Works the same way as the GCC -include
                      option.
    * exclude       : list, [], Give a source file or source file directory to
                      exclude from the check. This applies only to source files
                      so header files include by source files are not matched.
                      Directory name is matched to all parts of the path.
    * inconclusive  : bool, None, Allow that Cppcheck reports even though the
                      analysis is inconclusive.
                      There are false positives with this option. Each result
                      must be carefully investigated before you know if it is
                      good or bad.
    * inline_disable: bool, True, Enable inline suppressions. Use them by
                      placing one or more comments, like:
                      `// cppcheck-suppress warningId` on the lines before the
                      warning to suppress.
    * jobs          : int, None, Start <jobs> threads to do the checking
                      simultaneously.
    * load_average  : float, None, Specifies that no new threads should be
                      started if there are other threads running and the load
                      average at least <load>.
    * proglang      : str, None, Forces cppcheck to check all files as the
                      given language. Valid values are: c, c++.
    * library       : str, None, Load file <cfg> that contains information
                      about types and functions. With such information
                      Cppcheck understands your code better and therefore you
                      get better results. The std.cfg file that is distributed
                      with Cppcheck is loaded automatically. For more
                      information about library files, read the manual.
    * project       : str, None, Run Cppcheck on project. The <file> can be a
                      Visual Studio Solution (*.sln), Visual Studio Project
                      (*.vcxproj), or compile databse (compile_commands.json).
                      The files to analyse, include paths, defines, platform
                      and undefines in the specified file will be used.
    * max_configs   : int, None, Maximum number of configurations to check in
                      a file before skipping it. Default is '12'. If used
                      together with '--force', the last option is the one that
                      is effective.
    * platform      : str, None, Specifies platform specific types and sizes.
                      The available builtin platforms are:
                      * unix32
                        32 bit unix variant
                      * unix64
                        64 bit unix variant
                      * win32A
                        32 bit Windows ASCII character encoding
                      * win32W
                        32 bit Windows UNICODE character encoding
                      * win64
                        64 bit Windows
                      * native
                        Unspecified platform. Type sizes of host system are
                        assumed, but no further assumptions.
    * quiet         : bool, True, Do not show progress reports.
    * relative_paths: list, [], Use relative paths in output. When given,
                      <paths> are used as base. You can separate multiple
                      paths by ';'. Otherwise path where source files are
                      searched is used. We use string comparison to create
                      relative paths, so using e.g. ~ for home folder does not
                      work. It is currently only possible to apply the base
                      paths to files that are on a lower level in the
                      directory tree.
    * rule          : str, None, Match regular expression.
    * rule_file     : str, None, Use given rule file. For more information,
                      see:
                      http://sourceforge.net/projects/cppcheck/files/Articles/
    * standard      : list, None, Set standard. The available options are:
                      * posix
                        POSIX compatible code
                      * c89
                        C code is C89 compatible
                      * c99
                        C code is C99 compatible
                      * c11
                        C code is C11 compatible (default)
                      * c++03
                        C++ code is C++03 compatible
                      * c++11
                        C++ code is C++11 compatible (default)
    * disable       : list, None, Suppress warnings that match <spec>, The
                      format of <spec> is:
                      [error id]:[filename]:[line]
                      The [filename] and [line] are optional. If [error id] is
                      a wildcard '*', all error ids match.
    * disable_file  : str, None, Suppress warnings listed in the file. Each
                      suppression is in the same format as <spec> above.
    * template      : str, None, Format the error messages. E.g.
                      '{file}:{line},{severity},{id},{message}' or
                      '{file}({line}):({severity}) {message}' or
                      '{callstack} {message}'
                      Pre-defined templates: gcc, vs, edit.
    * verbose       : bool, None, Output more detailed error information.
    * xml           : bool, None, Write results in xml format to error stream
                      (stderr).
    * xml_version   : str, None, Select the XML file version. Currently
                      versions 1 and 2 are available. The default version is
                      1.

Requirements:

    * cppcheck
      to install, for example run `apt-get install cppcheck`

"""
from pybuildtool import BaseTask, expand_resource, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        args.append('--error-exitcode=-1')

        # Change current directory, before running cppcheck
        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)
            if self.workdir is None:
                self.bld.fatal(cfg['work_dir'] + ' not found.')

        # Build dir
        c = cfg.get('build_dir')
        if c:
            args.append('--cppcheck-build-dir=' + expand_resource(self.group,
                    c))

        # Check library
        c = cfg.get('check_library')
        if c:
            args.append('--check-library')

        # Config exclude directory
        for path in make_list(cfg.get('config_exclude')):
            args.append('--config-exclude=' + expand_resource(self.group, path))

        # Config excludes file
        c = cfg.get('config_exclude_file')
        if c:
            args.append('--config-excludes-file=' + expand_resource(self.group,
                    c))

        # Dump
        c = cfg.get('dump')
        if c:
            args.append('--dump')

        # Defines
        for define in make_list(cfg.get('define')):
            args.append('-D' + define)

        # Undefines
        for define in make_list(cfg.get('undefine')):
            args.append('-U' + define)

        # Enable additional checks
        c = make_list(cfg.get('enable_check'))
        if c:
            args.append('--enable=' + ','.join(c))

        # Exit code suppressions
        for path in make_list(cfg.get('ignore_files')):
            args.append('--exitcode-suppressions=' + expand_resource(self.group,
                    path))

        # Search directory for include files
        for path in make_list(cfg.get('include_path')):
            args.append('-I ' + expand_resource(self.group, path))

        # Include
        for path in make_list(cfg.get('include')):
            args.append('--include=' + expand_resource(self.group, path))

        # Exclude
        for path in make_list(cfg.get('exclude')):
            args.append('-i ' + expand_resource(self.group, path))

        # Inconclusive, allow false positives
        c = cfg.get('inconclusive')
        if c:
            args.append('--inconclusive')

        # Inline suppression
        c = cfg.get('inline_disable', True)
        if c:
            args.append('--inline-suppr')

        # Parallel
        c = cfg.get('jobs')
        if c:
            args.append('-j %i' % c)

        # Max processor usage
        c = cfg.get('load_average')
        if c:
            args.append('-l %f' % c)

        # Programming language
        c = cfg.get('proglang')
        if c:
            args.append('--language=' + c)

        # Library
        c = cfg.get('library')
        if c:
            args.append('--library=' + expand_resource(self.group, c))

        # Project file
        c = cfg.get('project')
        if c:
            args.append('--project=' + expand_resource(self.group, c))

        # Maximum configs
        c = cfg.get('max_configs')
        if c:
            args.append('--max-configs=%i' % c)

        # Platform
        c = cfg.get('platform')
        if c:
            args.append('--platform=' + c)

        # Quiet
        c = cfg.get('quiet', True)
        if c:
            args.append('--quiet')

        # Relative paths
        c = make_list(cfg.get('relative_paths'))
        if c:
            args.append('--relative-paths=' + ';'.join(c))

        # Rule
        c = cfg.get('rule')
        if c:
            args.append("--rule='%s'" % c)

        # Rule file
        c = cfg.get('rule_file')
        if c:
            args.append('--rule-file=' + expand_resource(self.group, c))

        # Standard
        for std in make_list(cfg.get('standard')):
            args.append('--std=' + std)

        # Suppress
        for dis in make_list(cfg.get('disable')):
            args.append('--suppress=' + dis)

        # Suppressions list
        c = cfg.get('disable_file')
        if c:
            args.append('--suppressions-list=' + expand_resource(self.group,
                    c))

        # Template
        c = cfg.get('template')
        if c:
            args.append("--template='%s'" % c)

        # Verbose
        c = cfg.get('verbose')
        if c:
            args.append('--verbose')

        # XML
        c = cfg.get('xml')
        if c:
            args.append('--xml')

        # XML version
        c = cfg.get('xml_version')
        if c:
            args.append('--xml-version=' + c)


    def perform(self):
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
                exe=executable,
                arg=' '.join(self.args),
                in_=' '.join(self.file_in),
            ),
            **kwargs)


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program(tool_name)[0]



from hashlib import md5
import os
import re
#-
from ..misc.path import expand_resource


def token_to_filename(token_name, bld):
    return os.path.join(bld.variant_dir, '.tokens',
            token_name.replace('/', '__'))


class Rule():

    def __init__(self, group, config, file_in, file_out, depend_in, extra_out):
        self.conf = config or {}
        self.file_in = file_in or []
        self.file_out = file_out or []
        self.depend_in = depend_in or []
        self.extra_out = extra_out or []
        self.group = group
        self.bld = group.context

        # expands wildcards (using ant_glob)
        for fs in (self.file_in, self.depend_in):
            self._expand_input_wilcards(fs)

        # normalize `_replace_patterns_`, must be a list of list
        self.conf.setdefault('_replace_patterns_', [])


    def _expand_input_wilcards(self, items):
        for_removal = []
        for_insertion = []
        for f in items:
            if not ('*' in f or '?' in f):
                continue
            for_removal.append(f)
            if os.path.isabs(f):
                paths = self.bld.root.ant_glob(f.lstrip('/'))
                for_insertion += (node.abspath() for node in paths)
            else:
                paths = self.bld.path.ant_glob(f)
                for_insertion += (node.relpath() for node in paths)
        for f in for_removal:
            items.remove(f)
        items += for_insertion


    def _extra_plus_token(self, file_out=None):
        for f in self.extra_out:
            yield f

        token_out = token_to_filename(self.group.get_name(), self.bld)
        if file_out:
            if hasattr(file_out, 'encode'):
                token_out += '-' + md5(file_out.encode()).hexdigest()
            else:
                token_out += '-' + md5(file_out).hexdigest()

        group_name = self.group.get_name()
        try:
            token_names = self.bld._token_names[group_name]
        except KeyError:
            token_names = []
            self.bld._token_names[group_name] = token_names
        except AttributeError:
            token_names = []
            self.bld._token_names = {group_name: token_names}
        token_names.append(token_out)

        yield token_out


    @property
    def files(self):
        # returns the output files after being processes by this tool
        result = []
        #if not self.file_out:
        #    return result

        for fo in self.file_out:
            is_dir = fo.endswith(os.path.sep)
            if is_dir:
                for fi in self.file_in:
                    fofi = fi
                    for (pat, rep) in self.conf['_replace_patterns_']:
                        fofi = re.sub(pat, rep, fofi)
                    basedir = self.conf.get('_source_basedir_', False)
                    if basedir:
                        basedir = expand_resource(self.group, basedir)
                    if basedir and fofi.startswith(basedir):
                        fofi = fofi[len(basedir):].strip('/')
                    else:
                        fofi = os.path.basename(fofi)
                    result.append(os.path.join(fo, fofi))
            else:
                result.append(fo)
        for fo in self.extra_out:
            result.append(fo)
        return result


    @property
    def rules(self):
        result = []

        if self.extra_out and (len(self.file_out) > 1 or\
                (self.file_out and self.file_out[0].endswith(
                os.path.sep))):

            self.bld.fatal('Cannot use extra_out with multiple file_out')

        for fo in self.file_out:
            if not self.file_in:
                # okay this is weird, no file_in but there is a file_out
                # it is possible though, but shouldn't you use extra_out?
                result.append({
                    'file_in': self.file_in,
                    'file_out': [fo],
                    'depend_in': self.depend_in,
                    'extra_out': self._extra_plus_token(fo),
                })
                continue

            if self.conf.get('_source_grouped_', False):
                result.append({
                    'file_in': self.file_in,
                    'file_out': [fo],
                    'depend_in': self.depend_in,
                    'extra_out': self._extra_plus_token(fo),
                })
                continue

            is_dir = fo.endswith(os.path.sep)
            for fi in self.file_in:
                if not is_dir:
                    result.append({
                        'file_in': [fi],
                        'file_out': [fo],
                        'depend_in': self.depend_in,
                        'extra_out': self._extra_plus_token(fo),
                    })
                    continue

                fofi = fi
                for (pat, rep) in self.conf['_replace_patterns_']:
                    fofi = re.sub(pat, rep, fofi)
                # use basedir to produce file_out
                basedir = self.conf.get('_source_basedir_', False)
                if basedir:
                    basedir = expand_resource(self.group, basedir)
                if basedir and fofi.startswith(basedir):
                    fofi = fofi[len(basedir):].strip('/')
                else:
                    fofi = os.path.basename(fofi)

                fofi = os.path.join(fo, fofi)
                result.append({
                    'file_in': [fi],
                    'file_out': [fofi],
                    'depend_in': self.depend_in,
                    'extra_out': self._extra_plus_token(fofi),
                })

        if not self.file_out:
            result.append({
                'file_in': self.file_in,
                'depend_in': self.depend_in,
                'extra_out': self._extra_plus_token(),
            })

        return result



"""
Compile jinja2 templates.

This tool will accept multiple file_in, but only the first one will be
processed, the others are treated as dependency.

Options:

    * search_dir    : str, [], directories to search for templates
    * context_python: str, [], python scripts containing context for template
                      if it has __init__.py in the same directory it will be
                      added to sys.path
                      it must have variable type dict named: `export`
    * context_yaml  : str, [], yaml file containing context for template
    * context_json  : str, [], json file containing context for template
    * context       : any, {}, context to be used in template

Requirements:

    * jinja2
      to install, run `pip install jinja2`

"""
from typing import Mapping, Sequence
from json import load as json_load
import os
import sys
from yaml import safe_load as yaml_load
from pybuildtool import BaseTask, expand_resource, make_list
from pybuildtool.misc.python import load_module_from_filename

tool_name = __name__


def dict_merge(destination, source):
    for k, v in source.items():
        if k not in destination:
            destination[k] = v
        elif isinstance(v, Mapping)\
                and isinstance(destination[k], Mapping):
            dict_merge(destination[k], v)
        elif isinstance(v, Sequence)\
                and isinstance(destination[k], Sequence):
            destination[k].extend(v)


class Task(BaseTask):

    name = tool_name
    search_dir = None
    context = {}

    def prepare(self):
        cfg = self.conf

        # Change current directory
        c = make_list(cfg.get('search_dir'))
        if c:
            c = [x for x in (expand_resource(self.group, f) for f\
                    in c) if x]
        if c:
            self.search_dir = c
        else:
            self.bld.fatal(('"search_dir" is required configuration '
                'for %s') % tool_name.capitalize())

        # Json context
        files = make_list(cfg.get('context_json'))
        if files:
            files = [x for x in (expand_resource(self.group, f) for f\
                    in files) if x]
        for json_file in files:
            with open(json_file, 'r', encoding='utf-8') as f:
                dict_merge(self.context, json_load(f))

        # Yaml context
        files = make_list(cfg.get('context_yaml'))
        if files:
            files = [x for x in (expand_resource(self.group, f) for f\
                    in files) if x]
        for yaml_file in files:
            with open(yaml_file, 'r', encoding='utf-8') as f:
                dict_merge(self.context, yaml_load(f))

        # Python context
        files = make_list(cfg.get('context_python'))
        if files:
            files = [x for x in (expand_resource(self.group, f) for f\
                    in files) if x]
        for python_file in files:
            dirname, filename = os.path.split(python_file)
            filebase, _ = os.path.splitext(filename)
            if os.path.exists(os.path.join(dirname, '__init__.py')):
                sys.path.append(dirname)
                mod = __import__(filebase)
            else:
                mod = load_module_from_filename(python_file, filebase)
            dict_merge(self.context, mod.export)

        dict_merge(self.context, cfg.get('context', {}))


    def perform(self):
        from jinja2 import Environment, FileSystemLoader # pylint:disable=import-error,import-outside-toplevel

        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        template_name = self.file_in[0]
        for s in self.search_dir:
            if not template_name.startswith(s):
                continue
            template_name = template_name[len(s):].strip('/')
            break
        else:
            self.bld.fatal('input for %s must be within `search_dir' %\
                    tool_name.capitalize())

        loader = FileSystemLoader(self.search_dir)
        env = Environment(loader=loader)
        template = env.get_template(template_name)
        rendered = template.render(self.context)
        for out in self.file_out:
            with open(out, 'w', encoding='utf-8') as f:
                f.write(rendered)
        return 0



"""
pngcrush is a png compressor

Options:

    * already       : int,   None
                      Already_crushed_size [e.g., 8192]
    * bail          : bool,  False
                      Bail out of trial when size exceeds best size found
    * bit_depth     : int,   None
                      Bit depth to use in output file
    * blacken       : bool,  False
                      Zero samples underlying fully-transparent pixels
    * brute         : int,   148
                      Use brute-force: try 138 different methods [11-148]
    * color_type    : int,   None
                      Color_type of output file [0, 2, 4, or 6]
    * double_gamma  : bool,  False
                      Used for fixing gamma in PhotoShop 5.0/5.02 files
    * filter        : int,   None
                      user_filter [0-5] for specified method
    * fix           : bool,  True
                      Fix otherwise fatal conditions such as bad CRCs
    * force         : bool,  False
                      Write a new output file even if larger than input
    * gamma         : float, None
                      Gamma (float or fixed*100000, e.g., 0.45455 or 45455)
    * huffman       : bool,  False
                      Use only zlib strategy 2, Huffman-only
    * iccp          : int,   None
                      Length "Profile Name" iccp_file
    * itxt          : str,   None
                      b[efore_IDAT]|a[fter_IDAT] "keyword"
    * keep          : bool,  False
                      chunk_name
    * level         : int,   None
                      zlib_compression_level [0-9] for specified method
    * loco          : bool,  False
                      "loco crush" truecolor PNGs
    * method        : int,   None
                      method [1 through 200]
    * max           : int,   None
                      maximum_IDAT_size [default 8192]
    * mng           : bool,  False
                      Write a new MNG, do not crush embedded PNGs
    * nobail        : bool,  False
                      Do not bail out early from trial -- see "-bail"
    * nofilecheck   : bool,  False
                      Do not check for infile.png == outfile.png
    * nolimits      : bool,  False
                      Turns off limits on width, height, cache, malloc
    * noreduce      : bool,  False
                      Turns off "-reduce" operations
    * oldtimestamp  : bool,  False
                      Do not reset file modification time
    * reduce        : bool,  False
                      Do lossless color-type or bit-depth reduction
    * rem           : bool,  False
                      chunkname (or "alla" or "allb")
    * replace_gamma : float, None
                      Gamma (float or fixed*100000) even if it is present
    * resolution    : int,   None
                      Resolution in dpi
    * rle           : bool,  False
                      Use only zlib strategy 3, RLE-only
    * save          : bool,  False
                      Keep all copy-unsafe PNG chunks
    * srgb          : int,   None
                      [0, 1, 2, or 3]
    * ster          : int,   None
                      [0 or 1]
    * text          : str,   None
                      b[efore_IDAT]|a[fter_IDAT] "keyword" "text"
    * trns_array    : str,   None
                      n trns[0] trns[1] .. trns[n-1]
    * trns          : str,   None
                      index red green blue gray
    * window_size   : int,   None
                      compression_window_size [32, 16, 8, 4, 2, 1, 512]
    * zlib          : int,   None
                      zlib_strategy [0, 1, 2, or 3] for specified method
    * zmem          : int,   None
                      zlib_compression_mem_level [1-9, default 9]
    * zitxt         : str,   None
                      b|a "keyword" "lcode" "tkey" "text"
    * ztxt          : str,   None
                      b[efore_IDAT]|a[fter_IDAT] "keywrod" "text"
    * quiet         : bool,  True
                      quiet

Requirements:

    * pngcrush
      to install, for example run `apt-get install pngcrush`

"""
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        # already_crushed_size [e.g., 8192]
        if cfg.get('already', None):
            args.append('-already=%i' % cfg['already'])

        # bail out of trial when size exceeds best size found
        if cfg.get('bail', False):
            args.append('-bail')

        # bit depth to use in output file
        if cfg.get('bit_depth', None):
            args.append('-bit_depth=%i' % cfg['bit_depth'])

        # zero samples underlying fully-transparent pixels
        if cfg.get('blacken', False):
            args.append('-blacken')

        # use brute-force: try 138 different methods [11-148]
        args.append('-brute=%i' % cfg.get('brute', 148))

        # color_type of output file [0, 2, 4, or 6]
        if cfg.get('color_type', None):
            args.append('-c=%i' % cfg['color_type'])

        # used for fixing gamma in PhotoShop 5.0/5.02 files
        if cfg.get('double_gamma', False):
            args.append('-double_gamma')

        # user_filter [0-5] for specified method
        if cfg.get('filter', None):
            args.append('-f=%i' % cfg['filter'])

        # fix otherwise fatal conditions such as bad CRCs
        if cfg.get('fix', True):
            args.append('-fix')

        # write a new output file even if larger than input
        if cfg.get('force', None):
            args.append('-force')

        # gamma (float or fixed*100000, e.g., 0.45455 or 45455)
        if cfg.get('gamma', None):
            args.append('-g=%s' % cfg['gamma'])

        # use only zlib strategy 2, Huffman-only
        if cfg.get('huffman', False):
            args.append('-huffman')

        # length "Profile Name" iccp_file
        if cfg.get('iccp', None):
            args.append('-iccp=%i' % cfg['iccp'])

        # b[efore_IDAT]|a[fter_IDAT] "keyword"
        if cfg.get('itxt', None):
            args.append('-itxt="%s"' % cfg['itxt'])

        # chunk_name
        if cfg.get('keep', False):
            args.append('-keep')

        # zlib_compression_level [0-9] for specified method
        if cfg.get('level', None):
            args.append('-l=%i' % cfg['level'])

        # "loco crush" truecolor PNGs
        if cfg.get('loco', False):
            args.append('-loco')

        # method [1 through 200]
        if cfg.get('method', None):
            args.append('-m=%i' % cfg['method'])

        # maximum_IDAT_size [default 8192]
        if cfg.get('max', None):
            args.append('-max=%i' % cfg['max'])

        # write a new MNG, do not crush embedded PNGs
        if cfg.get('mng', False):
            args.append('-mng')

        # do not bail out early from trial -- see "-bail"
        if cfg.get('nobail', None):
            args.append('-nobail')

        # do not check for infile.png == outfile.png
        if cfg.get('nofilecheck', False):
            args.append('-nofilecheck')

        # turns off limits on width, height, cache, malloc
        if cfg.get('nolimits', False):
            args.append('-nolimits')

        # turns off "-reduce" operations
        if cfg.get('noreduce', False):
            args.append('-noreduce')

        # do not reset file modification time
        if cfg.get('oldtimestamp', False):
            args.append('-oldtimestamp')

        # do lossless color-type or bit-depth reduction
        if cfg.get('reduce', True):
            args.append('-reduce')

        # chunkname (or "alla" or "allb")
        if cfg.get('rem', None):
            args.append('-rem="%s"' % cfg['rem'])

        # gamma (float or fixed*100000) even if it is present
        if cfg.get('replace_gamma', None):
            args.append('-replace_gamma=%s' % cfg['replace_gamma'])

        # resolution in dpi
        if cfg.get('resolution', None):
            args.append('-res=%i' % cfg['resolution'])

        # use only zlib strategy 3, RLE-only
        if cfg.get('rle', False):
            args.append('-rle')

        # keep all copy-unsafe PNG chunks
        if cfg.get('save', False):
            args.append('-save')

        # srgb
        if cfg.get('srgb', None):
            args.append('-srgb=%i' % cfg['srgb'])

        # ster
        if cfg.get('ster', None):
            args.append('-ster=%i' % cfg['ster'])

        # b[efore_IDAT]|a[fter_IDAT] "keyword" "text"
        if cfg.get('text', None):
            args.append('-text="%s"' % cfg['text'])

        # trns_array: n trns[0] trns[1] .. trns[n-1]
        if cfg.get('trns_array', None):
            args.append('-trns_array="%s"' % cfg['trns_array'])

        # index red green blue gray
        if cfg.get('trns', None):
            args.append('-trns="%s"' % cfg['trns'])

        # compression_window_size [32, 16, 8, 4, 2, 1, 512]
        if cfg.get('window_size', None):
            args.append('-w=%i' % cfg['window_size'])

        # zlib_strategy [0, 1, 2, or 3] for specified method
        if cfg.get('zlib', None):
            args.append('-z=%i' % cfg['zlib'])

        # zlib_compression_mem_level [1-9, default 9]
        if cfg.get('zmem', None):
            args.append('-zmem=%i' % cfg['zmem'])

        # b|a "keyword" "lcode" "tkey" "text"
        if cfg.get('zitxt', None):
            args.append('-zitxt="%s"' % cfg['zitxt'])

        # b[efore_IDAT]|a[fter_IDAT] "keywrod" "text"
        if cfg.get('ztxt', None):
            args.append('-ztxt="%s"' % cfg['ztxt'])

        # quiet
        if cfg.get('quiet', True):
            args.append('-q')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output' %\
                    tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program('pngcrush')[0]



from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_loader

def load_module_from_filename(filename, name):
    spec = spec_from_loader(name, SourceFileLoader(name, filename))
    mod = module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod



r"""
FlatBuffers is an efficient cross platform serialization library for C++, C#,
C, Go, Java, JavaScript, PHP, and Python. It was originally created at Google
for game development and other performance-critical applications.

Options:

    * binary : bool, None
             : Generate wire format binaries for any data definitions.

    * json : bool, None
           : Generate text output for any data definitions.

    * cpp : bool, None
          : Generate C++ headers for tables/structs.

    * go : bool, None
         : Generate Go files for tables/structs.

    * java : bool, None
           : Generate Java classes for tables/structs.

    * js : bool, None
         : Generate JavaScript code for tables/structs.

    * csharp : bool, None
             : Generate C# classes for tables/structs.

    * python : bool, None
             : Generate Python files for tables/structs.

    * php : bool, None
          : Generate PHP files for tables/structs.

    * output_dir : str, None, required
                 : Prefix PATH to all generated files.

    * include_dir : str, None
                  : Search for includes in the specified path.

    * strict_json : bool, None
                  : Strict JSON: field names must be / will be quoted,
                  : no trailing commas in tables/vectors.

    * allow_non_utf8 : bool, None
                     : Pass non-UTF-8 input through parser and emit nonstandard
                     : \x escapes in JSON. (Default is to raise parse error on
                     : non-UTF-8 input.)

    * defaults_json : bool, None
                    : Output fields whose value is the default when
                    :  writing JSON

    * unknown_json : bool, None
                   : Allow fields in JSON that are not defined in the
                   : schema. These fields will be discared when generating
                   : binaries.

    * no_prefix : bool, None
                : Don't prefix enum values with the enum type in C++.

    * scoped_enums : bool, None
                   : Use C++11 style scoped and strongly typed enums.
                   : also implies --no-prefix.

    * no_includes : bool, None
                  : Don't generate include statements for included
                  :  schemas the generated file depends on (C++).

    * gen_mutable : bool, None
                  : Generate accessors that can mutate buffers in-place.

    * gen_onefile : bool, None
                  : Generate single output file for C#.

    * gen_name_strings : bool, None
                       : Generate type name functions for C++.

    * escape_proto_ids : bool, None
                       : Disable appending '_' in namespaces names.

    * gen_object_api : bool, None
                     : Generate an additional object-based API.

    * cpp_ptr_type : str, None
                   : Set object API pointer type (default std::unique_ptr)

    * no_js_exports : bool, None
                    : Removes Node.js style export lines in JS.

    * goog_js_export : bool, None
                     : Uses goog.exports* for closure compiler exporting in JS.

    * raw_binary : bool, None
                 : Allow binaries without file_indentifier to be read.

This may crash flatc given a mismatched schema.

    * proto : bool, None
            : Input is a .proto, translate to .fbs.

    * grpc : bool, None
           : Generate GRPC interfaces for the specified languages

    * schema : bool, None
             : Serialize schemas instead of JSON (use with -b)

    * bfbs_comments : bool, None
                    : Add doc comments to the binary schema files.

    * conform : str, None
              : Specify a schema the following schemas should be
              : an evolution of. Gives errors if not.

    * conform_includes : str, None
                       : Include path for the schema given with --conform PATH

    * include_prefix : str, None
                     : Prefix this path to any generated include statements.

Requirements:

    * flatc
      to install, for example run `apt-get install flatbuffers`

"""
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        self.add_bool_args('binary', 'json', 'cpp', 'go', 'java', 'js',
                'csharp', 'python', 'php', 'strict_json', 'allow_non_utf8',
                'defaults_json', 'unknown_json', 'no_prefix', 'scoped_enums',
                'no_includes', 'gen_mutable', 'gen_onefile', 'gen_name_strings',
                'escape_proto_ids', 'gen_object_api', 'no_js_exports',
                'goog_js_export', 'raw_binary', 'proto', 'grpc', 'schema',
                'bfbs_comments')

        self.add_path_args('conform', 'conform_includes', 'include_prefix')

        self.add_str_args('cpp_ptr_type')

        c = cfg.get('output_dir')
        if c is None:
            self.bld.fatal('Option output_dir is required.')
        else:
            output_dir = expand_resource(self.group, c)
            if output_dir is None:
                self.bld.fatal(c + ' was not found.')
            args.append('-o')
            args.append(output_dir)

        c = cfg.get('include_dir')
        if c:
            include_dir = expand_resource(self.group, c)
            if include_dir is None:
                self.bld.fatal(c + ' was not found.')
            args.append('-I')
            args.append(include_dir)


    def perform(self):
        if not self.file_in:
            self.bld.fatal('%s needs input' % tool_name.capitalize())
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg}'.format(
                exe=executable,
                arg=' '.join(self.args),
            ))


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program('flatc')[0]



""" Copy files.
"""
from shutil import copyfile, Error
from pybuildtool import BaseTask
from pybuildtool.misc.resource import get_filehash

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_noop_retcodes_': 666,
    }
    name = tool_name

    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input, got %s' % (
                    tool_name.capitalize(), repr(self.file_in)))

        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output, got %s' % (
                    tool_name.capitalize(), repr(self.file_out)))

        try:
            source_hash = get_filehash(self.file_in[0])
            if source_hash and source_hash == get_filehash(self.file_out[0]):
                return 666
            copyfile(self.file_in[0], self.file_out[0])
            return 0
        except Error:
            self.bld.fatal('tried to copy file to itself')
        except IOError:
            self.bld.fatal('destination location cannot be written')
        return 1



"""
Browserify lets you require('modules') in the browser by bundling up all of
your dependencies.

This tool will accept multiple file_in, but only the first one will be
processed, the others are treated as dependency.

Options:

    * transform_module: list, [], use a transform module on
                        top-level files

Requirements:

    * node.js
    * browserify
      to install, run `npm install --save-dev browserify`

"""

import os
from pybuildtool import BaseTask, make_list

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_source_grouped_': True,
    }
    name = tool_name

    def prepare(self):
        args = self.args
        conf = self.conf

        for mod in make_list(conf.get('transform_module')):
            args.append("--transform '%s'" % mod)


    def perform(self):
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} -o {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    bin_path = 'node_modules/browserify/bin/cmd.js'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('browserify')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



# pylint:disable=line-too-long
r'''
Run MSBuild.

Builds the specified targets in the project file. If a project file is not
specified, MSBuild searches the current working directory for a file that has
a file extension that ends in "proj" and uses that file.

Options:

    * target : list, None
             : Build these targets in this project.
             : Example:
             :   - Resources
             :   - Compile

    * property : list, None
               : Set or override these project-level properties. <n> is the
               : property name, and <v> is the property value.
               : Example:
               :   - WarningLevel=2
               :   - OutDir=bin\Debug\

    * max_cpu_count : int, None
                    : Specifies the maximum number of concurrent processes to
                    : build with. If the switch is not used, the default value
                    : used is 1. If the switch is used without a value MSBuild
                    : will use up to the number of processors on the computer.

    * tools_version : str, None
                    : The version of the MSBuild Toolset (tasks, targets, etc.)
                    : to use during build. This version will override the
                    : versions specified by individual projects.
                    : Example: 3.5

    * verbosity : str, None
                : Display this amount of information in the event log.
                : The available verbosity levels are:
                :   - q[uiet]
                :   - m[inimal]
                :   - n[ormal]
                :   - d[etailed]
                :   - diag[nostic]

    * console_logger_parameters : list, None
                                : Parameters to console logger.
                                : The available parameters are:
                                :   - PerformanceSummary
                                :     Show time spent in tasks, targets and
                                :     projects.
                                :   - Summary
                                :     Show error and warning summary at the end.
                                :   - NoSummary
                                :     Don't show error and warning summary at
                                :     the end.
                                :   - ErrorsOnly
                                :     Show only errors.
                                :   - WarningsOnly
                                :     Show only warnings.
                                :   - NoItemAndPropertyList
                                :     Don't show list of items and properties at
                                :     the start of each project build.
                                :   - ShowCommandLine
                                :     Show TaskCommandLineEvent messages
                                :   - ShowTimestamp
                                :     Display the Timestamp as a prefix to any
                                :     message.
                                :   - ShowEventId
                                :     Show eventId for started events, finished
                                :     events, and messages
                                :   - ForceNoAlign
                                :     Does not align the text to the size of
                                :     the console buffer
                                :   - DisableConsoleColor
                                :     Use the default console colors for all
                                :     logging messages.
                                :   - DisableMPLogging
                                :     Disable the multiprocessor logging style
                                :     of output when running in
                                :     non-multiprocessor mode.
                                :   - EnableMPLogging
                                :     Enable the multiprocessor logging style
                                :     even when running in non-multiprocessor
                                :     mode. This logging style is on by default.
                                :   - Verbosity
                                :     overrides the /verbosity setting for this
                                :     logger.
                                : Example:
                                :   - PerformanceSummary
                                :   - NoSummary
                                :   - Verbosity=minimal

    * no_console_logger : bool, None
                        : Disable the default console logger and do not log
                        : events to the console.

    * file_logger[n] : bool, None
                     : Logs the build output to a file. By default the file is
                     : in the current directory and named "msbuild[n].log".
                     : Events from all nodes are combined into a single log.
                     : The location of the file and other parameters for the
                     : fileLogger can be specified through the addition of the
                     : "/fileLoggerParameters[n]" switch.
                     : "n" if present can be a digit from 1-9, allowing up to
                     : 10 file loggers to be attached. (Short form: /fl[n])

    * file_logger_parameters[n] : list, None
                                : Provides any extra parameters for file
                                : loggers.
                                : The presence of this switch implies the
                                : corresponding /filelogger[n] switch.
                                : "n" if present can be a digit from 1-9.
                                : /fileloggerparameters is also used by any
                                : distributed file logger, see description of
                                : /distributedFileLogger.
                                :
                                : The same parameters listed for the console
                                : logger are available. Some additional
                                : available parameters are:
                                :   - LogFile
                                :     path to the log file into which the build
                                :     log will be written.
                                :   - Append
                                :     determines if the build log will be
                                :     appended to or overwrite the log file.
                                :     Setting the switch appends the build log
                                :     to the log file;
                                :     Not setting the switch overwrites the
                                :     contents of an existing log file.
                                :     The default is not to append to the log
                                :     file.
                                :   - Encoding
                                :     specifies the encoding for the file, for
                                :     example, UTF-8, Unicode, or ASCII
                                :
                                : Default verbosity is Detailed.
                                : Examples:
                                :   - LogFile=MyLog.log
                                :   - Append
                                :   - Verbosity=diagnostic
                                :   - Encoding=UTF-8

    * distributed_logger : list, None
                         : Use this logger to log events from MSBuild, attaching
                         : a different logger instance to each node. To specify
                         : multiple loggers, specify each logger separately.
                         : The <logger> syntax is:
                         : [<logger class>,]<logger assembly>[;<logger parameters>]  #  noqa
                         : The <logger class> syntax is:
                         : [<partial or full namespace>.]<logger class name>
                         : The <logger assembly> syntax is:
                         : {<assembly name>[,<strong name>] | <assembly file>}
                         : The <logger parameters> are optional, and are passed
                         : to the logger exactly as you typed them.
                         : Examples:
                         :   - XMLLogger,MyLogger,Version=1.0.2,Culture=neutral
                         :   - MyLogger,C:\My.dll*ForwardingLogger,C:\Logger.dll

    * distributed_file_logger : bool, None
                              : Logs the build output to multiple log files, one
                              : log file per MSBuild node. The initial location
                              : for these files is the current directory. By
                              : default the files are called
                              : "MSBuild<nodeid>.log". The location of the files
                              : and other parameters for the fileLogger can be
                              : specified with the addition of the
                              : "/fileLoggerParameters" switch.
                              :
                              : If a log file name is set through the
                              : fileLoggerParameters switch the distributed
                              : logger will use the fileName as a template and
                              : append the node id to this fileName to create a
                              : log file for each node.

    * logger : list, None
             : Use this logger to log events from MSBuild. To specify multiple
             : loggers, specify each logger separately.
             : The <logger> syntax is:
             : [<logger class>,]<logger assembly>[;<logger parameters>]
             : The <logger class> syntax is:
             : [<partial or full namespace>.]<logger class name>
             : The <logger assembly> syntax is:
             : {<assembly name>[,<strong name>] | <assembly file>}
             : The <logger parameters> are optional, and are passed to the
             : logger exactly as you typed them.
             : Examples:
             :   - XMLLogger,MyLogger,Version=1.0.2,Culture=neutral
             :   - XMLLogger,C:\Loggers\MyLogger.dll;OutputAsHTML

    * validate : str, None
               : Validate the project against the default schema or against the
               : specified schema.
               : Example: MyExtendedBuildSchema.xsd

    * ignore_project_extensions : list, None
                                : List of extensions to ignore when determining
                                : which project file to build.
                                : Example:
                                :   - .sln

    * node_reuse : bool, None
                 : Enables or Disables the reuse of MSBuild nodes.
                 : The parameters are:
                 :   - True
                 :     Nodes will remain after the build completes and will be
                 :     reused by subsequent builds (default)
                 :   - False
                 :     Nodes will not remain after the build completes

    * preprocess : str, None
                 : Creates a single, aggregated project file by inlining all the
                 : files that would be imported during a build, with their
                 : boundaries marked. This can be useful for figuring out what
                 : files are being imported and from where, and what they will
                 : contribute to the build. By default the output is written to
                 : the console window. If the path to an output file is provided
                 : that will be used instead.
                 : Example: out.txt

    * detailed_summary : bool, None
                       : Shows detailed information at the end of the build
                       : about the configurations built and how they were
                       : scheduled to nodes.

    * response_file : list, None
                    : Insert command-line settings from a text file.

    * no_auto_response : bool, None
                       : Do not auto-include any MSBuild.rsp files.

    * no_logo : bool, None
              : Do not display the startup banner and copyright message.

Requirements:

    * Microsoft Visual Studio

'''
# pylint:enable=line-too-long
import os
from pybuildtool import BaseTask, make_list, PATH

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        cfg = self.conf
        arg = self.args

        targets = make_list(cfg.get('target'))
        if targets:
            arg.append('/target:' + ';'.join(targets))

        properties = make_list(cfg.get('property'))
        if properties:
            arg.append('/property:' + ';'.join(properties))

        c = cfg.get('max_cpu_count')
        if c is not None:
            if not c:
                arg.append('/maxcpucount')
            else:
                arg.append('/maxcpucount:%i' % int(c))

        c = cfg.get('tools_version')
        if c:
            arg.append('/toolsversion:' + c)

        c = cfg.get('verbosity')
        if c:
            arg.append('/verbosity:' + c)

        params = make_list(cfg.get('console_logger_parameters'))
        if params:
            arg.append('/consoleloggerparameters:' + ';'.join(params))

        c = cfg.get('no_console_logger')
        if c:
            arg.append('/noconsolelogger')

        c = cfg.get('file_logger')
        if c:
            arg.append('/fileLogger')

        for n in range(1, 10):
            c = cfg.get('file_logger%i' % n)
            if c:
                arg.append('/fileLogger%i' % n)

        params = make_list(cfg.get('file_logger_parameters'))
        if params:
            arg.append('/fileloggerparameters:' + ';'.join(params))

        for n in range(1, 10):
            params = make_list(cfg.get('file_logger_parameters%i' % n))
            if params:
                arg.append('/fileloggerparameters%i:%s' % (n, ';'.join(params)))

        loggers = make_list(cfg.get('distributed_logger'))
        for logger in loggers:
            arg.append('/distributedlogger:' + logger)

        c = cfg.get('distributed_file_logger')
        if c:
            arg.append('/distributedFileLogger')

        loggers = make_list(cfg.get('logger'))
        for logger in loggers:
            arg.append('/logger:' + logger)

        c = cfg.get('validate')
        if c is not None:
            if c:
                arg.append('/validate:' + c)
            else:
                arg.append('/validate')

        c = make_list(cfg.get('ignore_project_extensions'))
        if c:
            arg.append('/ignoreprojectextensions:' + ';'.join(c))

        c = cfg.get('node_reuse')
        if c is not None:
            if c:
                arg.append('/nodeReuse:true')
            else:
                arg.append('/nodeReuse:false')

        c = cfg.get('preprocess')
        if c:
            arg.append('/preprocess:' + c)

        c = cfg.get('detailed_summary')
        if c:
            arg.append('/detailedsummary')

        files = make_list(cfg.get('response_file'))
        for file_ in files:
            arg.append('@' + file_)

        c = cfg.get('no_auto_response')
        if c:
            arg.append('/noautoresponse')

        c = cfg.get('no_logo')
        if c:
            arg.append('/nologo')


    def perform(self):
        executable = self.env['%s_BIN' % tool_name.upper()]
        cmds = [executable] + self.args + self.file_in
        return self.exec_command(cmds)


def configure(conf):
    bin_path = PATH(r'C:\Program Files (x86)\MSBuild\14.0\Bin\MSBuild.exe')
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('MSBuild.exe')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Lint reStructuredText files.

Options:

    * encoding         : str, 'utf-8'
                         File encoding, argument to python open()
    * ignore_directives: list, []
                         Ignore errors about unknown directives.
    * ignore_roles     : list, []
                         Ignore errors about unknown roles.

Requirements:

    * pygments
      to install, run `pip install pygments`
    * restructuredtext-lint
      to install, run `pip install restructuredtext-lint`

"""
import os
import re
from pybuildtool import BaseTask, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    encoding = None
    regex_ignores = None

    def prepare(self):
        cfg = self.conf

        # Encoding of the input file (e.g. utf-8)
        self.encoding = cfg.get('encoding', 'utf-8')

        self.regex_ignores = []

        # Ignore unknown directives
        unknown = make_list(cfg.get('ignore_directives'))
        if unknown:
            self.regex_ignores.append(re.compile('^Unknown directive type "(' +\
                    '|'.join(unknown) + ')".*'))

            self.regex_ignores.append(re.compile('^No directive entry for "(' +\
                    '|'.join(unknown) + ')".*'))

        # Ignore unknown roles
        unknown = make_list(cfg.get('ignore_roles'))
        if unknown:
            self.regex_ignores.append(re.compile('^Unknown interpreted text ' +\
                    'role "(' + '|'.join(unknown) + ')".*'))

            self.regex_ignores.append(re.compile('^No role entry for "(' +\
                    '|'.join(unknown) + ')".*'))


    def perform(self):
        from restructuredtext_lint import lint_file # pylint:disable=import-error,import-outside-toplevel

        result = 0
        for filename in self.file_in:
            errors = lint_file(filename, encoding=self.encoding)
            if not errors:
                continue

            relpath = os.path.relpath(filename)
            if relpath.startswith('.'):
                relpath = filename

            print_filename = True
            for error in errors:
                ignored = False
                for re_ignore in self.regex_ignores:
                    if re_ignore.match(error.message):
                        ignored = True
                        break

                if ignored:
                    continue

                if print_filename:
                    print('************* File ' + relpath)
                    print_filename = False

                if error.level > 2:
                    result += 1

                message = error.message.replace('\n', ' ')
                print('%s: %i: %s' % (error.type[0], error.line, message))

        return result


def configure(conf):
    conf.start_msg("Checking for python module '%s'" % tool_name)
    try:
        import restructuredtext_lint # pylint:disable=unused-import,unused-variable,import-outside-toplevel
    except ImportError:
        conf.end_msg('not found', color='YELLOW')



"""
very quick commandline wrapper around OpenTypeUtilities.cpp from Chromium,
used to make EOT (Embeddable Open Type) files from TTF (TrueType/OpenType
Font) files

Requirements:

    * ttf2eot
      to install, download/compile from https://github.com/metaflop/ttf2eot

"""
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_replace_patterns_': ((r'\.ttf$', '.eot'), (r'\.otf$', '.eot'))
    }
    name = tool_name

    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        ret = self.exec_command(
            '{exe} {arg} < {in_} > {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ), stdout=None)

        return ret


def configure(conf):
    bin_path = conf.find_program('ttf2eot')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



'''
Watch files for changes and run build

add option `browser-notifier`, see the tool `browser_reload_notifier`
'''
import signal
from waflib import Context # pylint:disable=import-error

from .application import Application
from .file_observer import FileObserver


def watch(bld):
    app = Application(bld)
    signal.signal(signal.SIGINT, app.stop)
    signal.signal(signal.SIGTERM, app.stop)

    print('We are WATCHING your every moves ...')
    app.run()


Context.g_module.__dict__['watch'] = watch



"""
rsync is a file transfer program capable of efficient remote update
via a fast differencing algorithm.

Options:

    * target : str, None
             : rsync destination

    * work_dir : str, None
               : change current directory before rsync

    * verbose : bool, None
              : increase verbosity

    * info : str, None
           : fine-grained informational verbosity

    * debug : str, None
            : fine-grained debug verbosity

    * msgs2stderr : bool, None
                  : special output handling for debugging

    * quiet : bool, None
            : suppress non-error messages

    * no_MOTD : bool, None
              : suppress daemon-mode MOTD (see manpage caveat)

    * checksum : bool, None
               : skip based on checksum, not mod-time & size

    * archive : bool, None
              : archive mode; equals -rlptgoD (no -H,-A,-X)

    * no_OPTION : list, None
                : turn off an implied OPTION (e.g. --no-D)

    * recursive : bool, None
                : recurse into directories

    * relative : bool, None
               : use relative path names

    * no_implied_dirs : bool, None
                      : don't send implied dirs with --relative

    * backup : bool, None
             : make backups (see --suffix & --backup-dir)

    * backup_dir : str, None
                 : make backups into hierarchy based in DIR

    * suffix : str, None
             : set backup suffix (default ~ w/o --backup-dir)

    * update : bool, None
             : skip files that are newer on the receiver

    * inplace : bool, None
              : update destination files in-place (SEE MAN PAGE)

    * append : bool, None
             : append data onto shorter files

    * append_verify : bool, None
                    : like --append, but with old data in file checksum

    * dirs : bool, None
           : transfer directories without recursing

    * links : bool, None
            : copy symlinks as symlinks

    * copy_links : bool, None
                 : transform symlink into referent file/dir

    * copy_unsafe_links : bool, None
                        : only "unsafe" symlinks are transformed

    * safe_links : bool, None
                 : ignore symlinks that point outside the source tree

    * munge_links : bool, None
                  : munge symlinks to make them safer (but unusable)

    * copy_dirlinks : bool, None
                    : transform symlink to a dir into referent dir

    * keep_dirlinks : bool, None
                    : treat symlinked dir on receiver as dir

    * hard_links : bool, None
                 : preserve hard links

    * perms : bool, None
            : preserve permissions

    * executability : bool, None
                    : preserve the file's executability

    * chmod : str, None
            : affect file and/or directory permissions

    * acls : bool, None
           : preserve ACLs (implies --perms)

    * xattrs : bool, None
             : preserve extended attributes

    * owner : bool, None
            : preserve owner (super-user only)

    * group : bool, None
            : preserve group

    * devices : bool, None
              : preserve device files (super-user only)

    * specials : bool, None
               : preserve special files

    * times : bool, None
            : preserve modification times

    * omit_dir_times : bool, None
                     : omit directories from --times

    * omit_link_times : bool, None
                      : omit symlinks from --times

    * super : bool, None
            : receiver attempts super-user activities

    * fake_super : bool, None
                 : store/recover privileged attrs using xattrs

    * sparse : bool, None
             : handle sparse files efficiently

    * preallocate : bool, None
                  : allocate dest files before writing them

    * dry_run : bool, None
              : perform a trial run with no changes made

    * whole_file : bool, None
                 : copy files whole (without delta-xfer algorithm)

    * one_file_system : bool, None
                      : don't cross filesystem boundaries

    * block_size : int, None
                 : force a fixed checksum block-size

    * rsh : str, None
          : specify the remote shell to use

    * rsync_path : str, None
                 : specify the rsync to run on the remote machine

    * existing : bool, None
               : skip creating new files on receiver

    * ignore_existing : bool, None
                      : skip updating files that already exist on receiver

    * remove_source_files : bool, None
                          : sender removes synchronized files (non-dirs)

    * delete : bool, None
             : delete extraneous files from destination dirs

    * delete_before : bool, None
                    : receiver deletes before transfer, not during

    * delete_during : bool, None
                    : receiver deletes during the transfer

    * delete_delay : bool, None
                   : find deletions during, delete after

    * delete_after : bool, None
                   : receiver deletes after transfer, not during

    * delete_excluded : bool, None
                      : also delete excluded files from destination dirs

    * ignore_missing_args : bool, None
                          : ignore missing source args without error

    * delete_missing_args : bool, None
                          : delete missing source args from destination

    * ignore_errors : bool, None
                    : delete even if there are I/O errors

    * force : bool, None
            : force deletion of directories even if not empty

    * max_delete : int, None
                 : don't delete more than NUM files

    * max_size : int, None
               : don't transfer any file larger than SIZE

    * min_size : int, None
               : don't transfer any file smaller than SIZE

    * partial : bool, None
              : keep partially transferred files

    * partial_dir : str, None
                  : put a partially transferred file into DIR

    * delay_updates : bool, None
                    : put all updated files into place at transfer's end

    * prune_empty_dirs : bool, None
                       : prune empty directory chains from the file-list

    * numeric_ids : bool, None
                  : don't map uid/gid values by user/group name

    * usermap : str, None
              : custom username mapping

    * groupmap : str, None
               : custom groupname mapping

    * chown : str, None
            : simple username/groupname mapping

    * timeout : int, None
              : set I/O timeout in seconds

    * contimeout : int, None
                 : set daemon connection timeout in seconds

    * ignore_times : bool, None
                   : don't skip files that match in size and mod-time

    * remote-option : str, None
                    : send OPTION to the remote side only

    * size_only : bool, None
                : skip files that match in size

    * modify_window : int, None
                    : compare mod-times with reduced accuracy

    * temp_dir : str, None
               : create temporary files in directory DIR

    * fuzzy : bool, None
            : find similar file for basis if no dest file

    * compare_dest : str, None
                   : also compare destination files relative to DIR

    * copy_dest : str, None
                : ... and include copies of unchanged files

    * link_dest : str, None
                : hardlink to files in DIR when unchanged

    * compress : bool, None
               : compress file data during the transfer

    * compress_level : int, None
                     : explicitly set compression level

    * skip_compress : list, None
                    : skip compressing files with a suffix in LIST

    * cvs_exclude : bool, None
                  : auto-ignore files the same way CVS does

    * filter : list, None
             : add a file-filtering RULE

    * exclude : list, None
              : exclude files matching PATTERN

    * exclude_from : str, None
                   : read exclude patterns from FILE

    * include : list, None
              : don't exclude files matching PATTERN

    * include_from : str, None
                   : read include patterns from FILE

    * files_from : str, None
                 : read list of source-file names from FILE

    * from0 : bool, None
            : all *-from/filter files are delimited by 0s

    * protect_args : bool, None
                   : no space-splitting; only wildcard special-chars

    * address : str, None
              : bind address for outgoing socket to daemon

    * port : int, None
           : specify double-colon alternate port number

    * sockopts : str, None
               : specify custom TCP options

    * blocking_io : bool, None
                  : use blocking I/O for the remote shell

    * stats : bool, None
            : give some file-transfer stats

    * 8_bit_output : bool, None
                   : leave high-bit chars unescaped in output

    * human_readable : bool, None
                     : output numbers in a human-readable format

    * progress : bool, None
               : show progress during transfer

    * itemize_changes : bool, None
                      : output a change-summary for all updates

    * out_format : str, None
                 : output updates using the specified FORMAT

    * log_file : str, None
               : log what we're doing to the specified FILE

    * log_file_format : str, None
                      : log updates using the specified FMT

    * password_file : str, None
                    : read daemon-access password from FILE

    * list_only : bool, None
                : list the files instead of copying them

    * bwlimit : int, None
              : limit socket I/O bandwidth

    * outbuf : str, None
             : set output buffering to None, Line, or Block

    * write_batch : str, None
                  : write a batched update to FILE

    * only_write_batch : str, None
                       : like --write-batch but w/o updating destination

    * read_batch : str, None
                 : read a batched update from FILE

    * protocol : int, None
               : force an older protocol version to be used

    * iconv : str, None
            : request charset conversion of filenames

    * checksum_seed : int, None
                    : set block/file checksum seed (advanced)

    * ipv4 : bool, None
           : prefer IPv4

    * ipv6 : bool, None
           : prefer IPv6


Requirements:

    * rsync
      to install, for example run `apt-get install rsync`

"""
import os
from pybuildtool import BaseTask, expand_resource, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    conf = {
        '_source_grouped_': True,
    }
    target = None
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        c = cfg.get('target')
        if c is None:
            self.bld.fatal('"target" is required by %s' % tool_name)
        else:
            self.target = c.format(**self.group.get_patterns())

        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)

        self.add_bool_args('verbose', 'msgs2stderr', 'quiet', 'checksum',
                'archive', 'recursive', 'relative', 'no_implied_dirs', 'backup',
                'update', 'inplace', 'append', 'append_verify', 'dirs', 'links',
                'copy_links', 'copy_unsafe_links', 'safe_links', 'munge_links',
                'copy_dirlinks', 'keep_dirlinks', 'hard_links', 'perms',
                'executability', 'acls', 'xattrs', 'owner', 'group', 'devices',
                'specials', 'times', 'omit_dir_times', 'omit_link_times',
                'super', 'fake_super', 'sparse', 'preallocate', 'dry_run',
                'whole_file', 'one_file_system', 'existing', 'ignore_existing',
                'remove_source_files', 'delete', 'delete_before',
                'delete_during', 'delete_delay', 'delete_after',
                'delete_excluded', 'ignore_missing_args', 'delete_missing_args',
                'ignore-errors', 'force', 'partial', 'delay_updates',
                'prune_empty_dirs', 'numeric_ids', 'ignore_times', 'size_only',
                'fuzzy', 'compress', 'cvs_exclude', 'from0', 'protect_args',
                'blocking_io', 'stats', '8_bit_output', 'human_readable',
                'progress', 'itemize_changes', 'list_only', 'ipv4', 'ipv6')

        self.add_int_args('block_size', 'max_delete', 'max_size', 'min_size',
                'timeout', 'contimeout', 'modify_window', 'compress_level',
                'port', 'bwlimit', 'protocol', 'checksum_seed')

        self.add_list_args_join(',', 'skip_compress')

        self.add_list_args_multi('filter', 'exclude', 'include')

        self.add_path_args('backup_dir', 'exclude_from', 'include_from',
                'files_from', 'password_file', 'read_batch')

        self.add_str_args('info', 'debug', 'suffix', 'chmod', 'rsh',
                'rsync_path', 'partial_dir', 'usermap', 'groupmap', 'chown',
                'remote_option', 'temp_dir', 'compare_dest', 'copy_dest',
                'link_dest', 'address', 'sockopts', 'out_format', 'log_file',
                'log_file_format', 'write_batch', 'only_write_batch', 'iconv')

        c = cfg.get('no_MOTD')
        if c:
            args.append('--no-motd')

        c = make_list(cfg.get('no_OPTION'))
        for option in c:
            args.append('--no-' + option)

        c = cfg.get('outbuf')
        if c:
            args.append('--outbuf=' + c[1].upper())


    def perform(self):
        if not self.file_in:
            self.bld.fatal('%s needs input' % tool_name.capitalize())
        if self.file_out:
            self.bld.fatal("%s doesn't produce output" % tool_name.capitalize())

        if self.conf.get('relative') and self.workdir:
            inputs = [os.path.relpath(x, self.workdir) for x in self.file_in]
        else:
            inputs = self.file_in

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            "{exe} {arg} {in_} {out}".format(
            exe=executable,
            arg=' '.join(self.args),
            in_=' '.join(inputs),
            out=self.target,
        ),
        **kwargs)


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program('rsync')[0]



""" Merge files from sources into copious targets.
"""
import os
from shutil import copyfileobj
from pybuildtool import BaseTask
from pybuildtool.misc.resource import get_filehash

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_source_grouped_': True,
        '_noop_retcodes_': 666,
    }
    name = tool_name

    def perform(self):
        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output' %\
                    tool_name.capitalize())

        if os.path.isdir(self.file_out[0]):
            self.bld.fatal('cannot concat to a directory')

        try:
            before_hash = get_filehash(self.file_out[0])
            with open(self.file_out[0], 'wb') as dest:
                for src_name in self.file_in:
                    with open(src_name, 'rb') as src:
                        copyfileobj(src, dest)

            if before_hash == get_filehash(self.file_out[0]):
                return 666
            return 0
        except OSError:
            return 1
